"""Subscribes to the ZMQ stats protocol and calls the stats event dispatcher when
we get stats from it."""


import zmq

import shinqlx
from threading import Thread


@shinqlx.next_frame
def dispatch_stats_event(stats):
    shinqlx.EVENT_DISPATCHERS["stats"].dispatch(stats)


@shinqlx.next_frame
def dispatch_game_start_event(data):
    shinqlx.EVENT_DISPATCHERS["game_start"].dispatch(data)


@shinqlx.next_frame
def dispatch_round_end_event(data):
    shinqlx.EVENT_DISPATCHERS["round_end"].dispatch(data)


@shinqlx.next_frame
def dispatch_game_end_event(data):
    # MATCH_REPORT event goes off with a map change and map_restart,
    # but we really only want it for when the game actually ends.
    # We use a variable instead of Game().state because by the
    # time we get the event, the game is probably gone.
    shinqlx.EVENT_DISPATCHERS["game_end"].dispatch(data)


@shinqlx.next_frame
def dispatch_player_death_event(data):
    # Dead player.
    sid = int(data["VICTIM"]["STEAM_ID"])
    player = (
        shinqlx.Plugin.player(sid)
        if sid
        else shinqlx.Plugin.player(data["VICTIM"]["NAME"])
    )

    # Killer player.
    if not data["KILLER"]:
        player_killer = None
    else:
        sid_killer = int(data["KILLER"]["STEAM_ID"])
        if sid_killer:
            player_killer = shinqlx.Plugin.player(sid_killer)
        else:  # It's a bot. Forced to use name as an identifier.
            player_killer = shinqlx.Plugin.player(data["KILLER"]["NAME"])

    shinqlx.EVENT_DISPATCHERS["death"].dispatch(player, player_killer, data)
    if player_killer:
        shinqlx.EVENT_DISPATCHERS["kill"].dispatch(player, player_killer, data)


@shinqlx.next_frame
def dispatch_team_switch_event(data):
    # No idea why they named it "KILLER" here, but whatever.
    player = shinqlx.Plugin.player(int(data["KILLER"]["STEAM_ID"]))
    if player is None:
        return
    old_team = data["KILLER"]["OLD_TEAM"].lower()
    new_team = data["KILLER"]["TEAM"].lower()
    if old_team != new_team:
        res = shinqlx.EVENT_DISPATCHERS["team_switch"].dispatch(
            player, old_team, new_team
        )
        if res is False:
            player.put(old_team)


class StatsListener(Thread):
    def __init__(self):
        super().__init__()

        self.done = False
        self._in_progress = False

        if not bool(int(shinqlx.get_cvar("zmq_stats_enable"))):
            self.done = True
            return

        host = shinqlx.get_cvar("zmq_stats_ip") or "127.0.0.1"
        port = shinqlx.get_cvar("zmq_stats_port") or shinqlx.get_cvar("net_port")
        self.address = f"tcp://{host}:{port}"
        self.password = shinqlx.get_cvar("zmq_stats_password")

    def keep_receiving(self):
        """Receives until 'self.done' is set to True."""
        self.start()

    def run(self):
        if self.done:
            return

        with zmq.Context().instance().socket(zmq.SUB) as socket:
            socket.setsockopt_string(zmq.PLAIN_USERNAME, "stats")
            socket.setsockopt_string(zmq.PLAIN_PASSWORD, self.password)
            socket.setsockopt_string(zmq.ZAP_DOMAIN, "stats")
            socket.connect(self.address)
            socket.subscribe("")

            poller = zmq.Poller()
            poller.register(socket, zmq.POLLIN)

            while True:  # Will throw an expcetion if no more data to get.
                pending_events = dict(poller.poll(timeout=250))
                for receiver in pending_events:
                    stats = receiver.recv_json()
                    dispatch_stats_event(stats)

                    if stats["TYPE"] == "MATCH_STARTED":
                        self._in_progress = True
                        dispatch_game_start_event(stats["DATA"])
                    elif stats["TYPE"] == "ROUND_OVER":
                        dispatch_round_end_event(stats["DATA"])
                    elif stats["TYPE"] == "MATCH_REPORT":
                        if self._in_progress:
                            dispatch_game_end_event(stats["DATA"])
                        self._in_progress = False
                    elif stats["TYPE"] == "PLAYER_DEATH":
                        dispatch_player_death_event(stats["DATA"])
                    elif stats["TYPE"] == "PLAYER_SWITCHTEAM":
                        dispatch_team_switch_event(stats["DATA"])

    def stop(self):
        self.done = True



import re

import shinqlx

_DUMMY_USERINFO = (
    "ui_singlePlayerActive\\0\\cg_autoAction\\1\\cg_autoHop\\0"
    "\\cg_predictItems\\1\\model\\bitterman/sport_blue\\headmodel\\crash/red"
    "\\handicap\\100\\cl_anonymous\\0\\color1\\4\\color2\\23\\sex\\male"
    "\\teamtask\\0\\rate\\25000\\country\\NO"
)


class NonexistentPlayerError(Exception):
    """An exception that is raised when a player that disconnected is being used
    as if the player were still present.

    """


class Player:
    """A class that represents a player on the server. As opposed to minqlbot,
    attributes are all the values from when the class was instantiated. This
    means for instance if a player is on the blue team when you check, but
    then moves to red, it will still be blue when you check a second time.
    To update it, use :meth:`~.Player.update`. Note that if you update it
    and the player has disconnected, it will raise a
    :exc:`shinqlx.NonexistentPlayerError` exception.

    """

    __slots__ = ("_valid", "_id", "_info", "_userinfo", "_steam_id", "_name")

    def __init__(self, client_id, info=None):
        self._valid = True

        # Can pass own info for efficiency when getting all players and to allow dummy players.
        if info:
            self._id = client_id
            self._info = info
        else:
            self._id = client_id
            self._info = shinqlx.player_info(client_id)
            if not self._info:
                self._invalidate(
                    f"Tried to initialize a Player instance of nonexistant player {client_id}."
                )

        self._userinfo = None
        self._steam_id = self._info.steam_id

        # When a player connects, the name field in the client struct has yet to be initialized,
        # so we fall back to the userinfo and try parse it ourselves to get the name if needed.
        if self._info.name:
            self._name = self._info.name
        else:
            self._userinfo = shinqlx.parse_variables(self._info.userinfo, ordered=True)
            self._name = self._userinfo.get("name", "")

    def __repr__(self):
        if not self._valid:
            return f"{self.__class__.__name__}(INVALID:'{self.clean_name}':{self.steam_id})"

        return (
            f"{self.__class__.__name__}({self._id}:'{self.clean_name}':{self.steam_id})"
        )

    def __str__(self):
        return self.name

    def __contains__(self, key):
        return key in self.cvars

    def __getitem__(self, key):
        return self.cvars[key]

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.steam_id == other.steam_id

        return self.steam_id == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def update(self):
        """Update the player information with the latest data. If the player
        disconnected it will raise an exception and invalidates a player.
        The player's name and Steam ID can still be accessed after being
        invalidated, but anything else will make it throw an exception too.

        :raises: shinqlx.NonexistentPlayerError

        """
        self._info = shinqlx.player_info(self._id)

        if not self._info or self._steam_id != self._info.steam_id:
            self._invalidate()
            return

        if self._info.name:
            self._name = self._info.name
        else:
            self._userinfo = shinqlx.parse_variables(self._info.userinfo, ordered=True)
            self._name = self._userinfo.get("name", "")

    def _invalidate(
        self, e="The player does not exist anymore. Did the player disconnect?"
    ):
        self._valid = False
        raise NonexistentPlayerError(e)

    @property
    def cvars(self):
        if not self._valid:
            self._invalidate()

        if not self._userinfo:
            self._userinfo = shinqlx.parse_variables(self._info.userinfo, ordered=True)

        return self._userinfo.copy()

    # noinspection PyUnresolvedReferences
    @cvars.setter
    def cvars(self, new_cvars):
        new = "".join([f"\\{key}\\{new_cvars[key]}" for key in new_cvars])
        shinqlx.client_command(self.id, f'userinfo "{new}"')

    @property
    def steam_id(self):
        return self._steam_id

    @property
    def id(self):
        return self._id

    @property
    def ip(self):
        if "ip" in self:
            return self["ip"].split(":")[0]
        return ""

    @property
    def clan(self):
        """The clan tag. Not actually supported by QL, but it used to be and
        fortunately the scoreboard still properly displays it if we manually
        set the configstring to use clan tags."""
        try:
            return shinqlx.parse_variables(shinqlx.get_configstring(529 + self._id))[
                "cn"
            ]
        except KeyError:
            return ""

    # noinspection PyUnresolvedReferences
    @clan.setter
    def clan(self, tag):
        index = self.id + 529
        cs = shinqlx.parse_variables(shinqlx.get_configstring(index), ordered=True)
        cs["xcn"] = tag
        cs["cn"] = tag
        new_cs = "".join([f"\\{key}\\{cs[key]}" for key in cs])
        shinqlx.set_configstring(index, new_cs)

    @property
    def name(self):
        return self._name + "^7"

    # noinspection PyUnresolvedReferences
    @name.setter
    def name(self, value):
        new = self.cvars
        new["name"] = value
        self.cvars = new

    @property
    def clean_name(self):
        """Removes color tags from the name."""
        return re.sub(r"\^\d", "", self.name)

    @property
    def qport(self):
        if "qport" in self:
            return int(self["qport"])

        return -1

    @property
    def team(self):
        return shinqlx.TEAMS[self._info.team]

    # noinspection PyUnresolvedReferences
    @team.setter
    def team(self, new_team):
        self.put(new_team)

    @property
    def colors(self):
        # Float because they can occasionally be floats for some reason.
        return float(self["color1"]), float(self["color2"])

    # noinspection PyUnresolvedReferences
    @colors.setter
    def colors(self, value):
        new = self.cvars
        c1, c2 = value
        new["color1"] = c1
        new["color2"] = c2
        self.cvars = new

    @property
    def model(self):
        return self["model"]

    # noinspection PyUnresolvedReferences
    @model.setter
    def model(self, value):
        new = self.cvars
        new["model"] = value
        self.cvars = new

    @property
    def headmodel(self):
        return self["headmodel"]

    # noinspection PyUnresolvedReferences
    @headmodel.setter
    def headmodel(self, value):
        new = self.cvars
        new["headmodel"] = value
        self.cvars = new

    @property
    def handicap(self):
        return self["handicap"]

    # noinspection PyUnresolvedReferences
    @handicap.setter
    def handicap(self, value):
        new = self.cvars
        new["handicap"] = value
        self.cvars = new

    @property
    def autohop(self):
        return bool(int(self["cg_autoHop"]))

    # noinspection PyUnresolvedReferences
    @autohop.setter
    def autohop(self, value):
        new = self.cvars
        new["autohop"] = int(value)
        self.cvars = new

    @property
    def autoaction(self):
        return bool(int(self["cg_autoAction"]))

    # noinspection PyUnresolvedReferences
    @autoaction.setter
    def autoaction(self, value):
        new = self.cvars
        new["cg_autoAction"] = int(value)
        self.cvars = new

    @property
    def predictitems(self):
        return bool(int(self["cg_predictItems"]))

    # noinspection PyUnresolvedReferences
    @predictitems.setter
    def predictitems(self, value):
        new = self.cvars
        new["cg_predictItems"] = int(value)
        self.cvars = new

    @property
    def connection_state(self):
        """A string describing the connection state of a player.

        Possible values:
        - *free* -- The player has disconnected and the slot is free to be used by someone else.
        - *zombie* -- The player disconnected and his/her slot will be available to other players shortly.
        - *connected* -- The player connected, but is currently loading the game.
        - *primed* -- The player was sent the necessary information to play, but has yet to send commands.
        - *active* -- The player finished loading and is actively sending commands to the server.

        In other words, if you need to make sure a player is in-game, check if ``player.connection_state == "active"``.

        """
        return shinqlx.CONNECTION_STATES[self._info.connection_state]

    @property
    def state(self):
        return shinqlx.player_state(self.id)

    @property
    def privileges(self):
        if self._info.privileges == shinqlx.PRIV_MOD:
            return "mod"
        if self._info.privileges == shinqlx.PRIV_ADMIN:
            return "admin"
        if self._info.privileges == shinqlx.PRIV_ROOT:
            return "root"
        if self._info.privileges == shinqlx.PRIV_BANNED:
            return "banned"
        return None

    # noinspection PyUnresolvedReferences
    @privileges.setter
    def privileges(self, value):
        if not value or value == "none":
            shinqlx.set_privileges(self.id, shinqlx.PRIV_NONE)
        elif value == "mod":
            shinqlx.set_privileges(self.id, shinqlx.PRIV_MOD)
        elif value == "admin":
            shinqlx.set_privileges(self.id, shinqlx.PRIV_ADMIN)
        else:
            raise ValueError("Invalid privilege level.")

    @property
    def country(self):
        return self["country"]

    # noinspection PyUnresolvedReferences
    @country.setter
    def country(self, value):
        new = self.cvars
        new["country"] = value
        self.cvars = new

    @property
    def valid(self):
        return self._valid

    @property
    def stats(self):
        return shinqlx.player_stats(self.id)

    @property
    def ping(self):
        return self.stats.ping

    def position(self, reset=False, **kwargs):
        pos = shinqlx.Vector3((0, 0, 0)) if reset else self.state.position

        if not kwargs:
            return pos

        x = pos.x if "x" not in kwargs else kwargs["x"]
        y = pos.y if "y" not in kwargs else kwargs["y"]
        z = pos.z if "z" not in kwargs else kwargs["z"]

        return shinqlx.set_position(self.id, shinqlx.Vector3((x, y, z)))

    def velocity(self, reset=False, **kwargs):
        vel = shinqlx.Vector3((0, 0, 0)) if reset else self.state.velocity

        if not kwargs:
            return vel

        x = vel.x if "x" not in kwargs else kwargs["x"]
        y = vel.y if "y" not in kwargs else kwargs["y"]
        z = vel.z if "z" not in kwargs else kwargs["z"]

        return shinqlx.set_velocity(self.id, shinqlx.Vector3((x, y, z)))

    def weapons(self, reset=False, **kwargs):
        weaps = shinqlx.Weapons((False,) * 15) if reset else self.state.weapons

        if not kwargs:
            return weaps

        g = weaps.g if "g" not in kwargs else kwargs["g"]
        mg = weaps.mg if "mg" not in kwargs else kwargs["mg"]
        sg = weaps.sg if "sg" not in kwargs else kwargs["sg"]
        gl = weaps.gl if "gl" not in kwargs else kwargs["gl"]
        rl = weaps.rl if "rl" not in kwargs else kwargs["rl"]
        lg = weaps.lg if "lg" not in kwargs else kwargs["lg"]
        rg = weaps.rg if "rg" not in kwargs else kwargs["rg"]
        pg = weaps.pg if "pg" not in kwargs else kwargs["pg"]
        bfg = weaps.bfg if "bfg" not in kwargs else kwargs["bfg"]
        gh = weaps.gh if "gh" not in kwargs else kwargs["gh"]
        ng = weaps.ng if "ng" not in kwargs else kwargs["ng"]
        pl = weaps.pl if "pl" not in kwargs else kwargs["pl"]
        cg = weaps.cg if "cg" not in kwargs else kwargs["cg"]
        hmg = weaps.hmg if "hmg" not in kwargs else kwargs["hmg"]
        hands = weaps.hands if "hands" not in kwargs else kwargs["hands"]

        return shinqlx.set_weapons(
            self.id,
            shinqlx.Weapons(
                (g, mg, sg, gl, rl, lg, rg, pg, bfg, gh, ng, pl, cg, hmg, hands)
            ),
        )

    def weapon(self, new_weapon=None):
        if new_weapon is None:
            return self.state.weapon
        if new_weapon in shinqlx.WEAPONS:
            pass
        elif new_weapon in shinqlx.WEAPONS.values():
            new_weapon = tuple(shinqlx.WEAPONS.values()).index(new_weapon)

        return shinqlx.set_weapon(self.id, new_weapon)

    def ammo(self, reset=False, **kwargs):
        a = shinqlx.Weapons((0,) * 15) if reset else self.state.ammo

        if not kwargs:
            return a

        g = a.g if "g" not in kwargs else kwargs["g"]
        mg = a.mg if "mg" not in kwargs else kwargs["mg"]
        sg = a.sg if "sg" not in kwargs else kwargs["sg"]
        gl = a.gl if "gl" not in kwargs else kwargs["gl"]
        rl = a.rl if "rl" not in kwargs else kwargs["rl"]
        lg = a.lg if "lg" not in kwargs else kwargs["lg"]
        rg = a.rg if "rg" not in kwargs else kwargs["rg"]
        pg = a.pg if "pg" not in kwargs else kwargs["pg"]
        bfg = a.bfg if "bfg" not in kwargs else kwargs["bfg"]
        gh = a.gh if "gh" not in kwargs else kwargs["gh"]
        ng = a.ng if "ng" not in kwargs else kwargs["ng"]
        pl = a.pl if "pl" not in kwargs else kwargs["pl"]
        cg = a.cg if "cg" not in kwargs else kwargs["cg"]
        hmg = a.hmg if "hmg" not in kwargs else kwargs["hmg"]
        hands = a.hands if "hands" not in kwargs else kwargs["hands"]

        return shinqlx.set_ammo(
            self.id,
            shinqlx.Weapons(
                (g, mg, sg, gl, rl, lg, rg, pg, bfg, gh, ng, pl, cg, hmg, hands)
            ),
        )

    def powerups(self, reset=False, **kwargs):
        pu = shinqlx.Powerups((0,) * 6) if reset else self.state.powerups

        if not kwargs:
            return pu

        quad = pu.quad if "quad" not in kwargs else round(kwargs["quad"] * 1000)
        bs = (
            pu.battlesuit
            if "battlesuit" not in kwargs
            else round(kwargs["battlesuit"] * 1000)
        )
        haste = pu.haste if "haste" not in kwargs else round(kwargs["haste"] * 1000)
        invis = (
            pu.invisibility
            if "invisibility" not in kwargs
            else round(kwargs["invisibility"] * 1000)
        )
        regen = (
            pu.regeneration
            if "regeneration" not in kwargs
            else round(kwargs["regeneration"] * 1000)
        )
        invul = (
            pu.invulnerability
            if "invulnerability" not in kwargs
            else round(kwargs["invulnerability"] * 1000)
        )

        return shinqlx.set_powerups(
            self.id, shinqlx.Powerups((quad, bs, haste, invis, regen, invul))
        )

    @property
    def holdable(self):
        return self.state.holdable

    # noinspection PyUnresolvedReferences
    @holdable.setter
    def holdable(self, value):
        if not value:
            shinqlx.set_holdable(self.id, 0)
        elif value == "teleporter":
            shinqlx.set_holdable(self.id, 27)
        elif value == "medkit":
            shinqlx.set_holdable(self.id, 28)
        elif value == "flight":
            shinqlx.set_holdable(self.id, 34)
            self.flight(reset=True)
        elif value == "kamikaze":
            shinqlx.set_holdable(self.id, 37)
        elif value == "portal":
            shinqlx.set_holdable(self.id, 38)
        elif value == "invulnerability":
            shinqlx.set_holdable(self.id, 39)
        else:
            raise ValueError("Invalid holdable item.")

    def drop_holdable(self):
        shinqlx.drop_holdable(self.id)

    def flight(self, reset=False, **kwargs):
        state = self.state
        if state.holdable != "flight":
            self.holdable = "flight"
            reset = True

        # Set to defaults on reset.
        fl = shinqlx.Flight((16000, 16000, 1200, 0)) if reset else state.flight

        fuel = fl.fuel if "fuel" not in kwargs else kwargs["fuel"]
        max_fuel = fl.max_fuel if "max_fuel" not in kwargs else kwargs["max_fuel"]
        thrust = fl.thrust if "thrust" not in kwargs else kwargs["thrust"]
        refuel = fl.refuel if "refuel" not in kwargs else kwargs["refuel"]

        return shinqlx.set_flight(
            self.id, shinqlx.Flight((fuel, max_fuel, thrust, refuel))
        )

    @property
    def noclip(self):
        return self.state.noclip

    # noinspection PyUnresolvedReferences
    @noclip.setter
    def noclip(self, value):
        shinqlx.noclip(self.id, bool(value))

    @property
    def health(self):
        return self.state.health

    # noinspection PyUnresolvedReferences
    @health.setter
    def health(self, value):
        shinqlx.set_health(self.id, value)

    @property
    def armor(self):
        return self.state.armor

    # noinspection PyUnresolvedReferences
    @armor.setter
    def armor(self, value):
        shinqlx.set_armor(self.id, value)

    @property
    def is_alive(self):
        return self.state.is_alive

    # noinspection PyUnresolvedReferences
    @is_alive.setter
    def is_alive(self, value):
        if not isinstance(value, bool):
            raise ValueError("is_alive needs to be a boolean.")

        cur = self.is_alive
        if cur and value is False:
            # TODO: Proper death and not just setting health to 0.
            self.health = 0
        elif not cur and value is True:
            shinqlx.player_spawn(self.id)

    @property
    def is_frozen(self):
        return self.state.is_frozen

    @property
    def score(self):
        return self.stats.score

    # noinspection PyUnresolvedReferences
    @score.setter
    def score(self, value):
        shinqlx.set_score(self.id, value)

    @property
    def channel(self):
        return shinqlx.TellChannel(self)

    def center_print(self, msg):
        shinqlx.send_server_command(self.id, f'cp "{msg}"')

    def tell(self, msg, **kwargs):
        return shinqlx.Plugin.tell(msg, self, **kwargs)

    def kick(self, reason=""):
        return shinqlx.Plugin.kick(self, reason)

    def ban(self):
        return shinqlx.Plugin.ban(self)

    def tempban(self):
        return shinqlx.Plugin.tempban(self)

    def addadmin(self):
        return shinqlx.Plugin.addadmin(self)

    def addmod(self):
        return shinqlx.Plugin.addmod(self)

    def demote(self):
        return shinqlx.Plugin.demote(self)

    def mute(self):
        return shinqlx.Plugin.mute(self)

    def unmute(self):
        return shinqlx.Plugin.unmute(self)

    def put(self, team):
        return shinqlx.Plugin.put(self, team)

    def addscore(self, score):
        return shinqlx.Plugin.addscore(self, score)

    def switch(self, other_player):
        return shinqlx.Plugin.switch(self, other_player)

    def slap(self, damage=0):
        return shinqlx.Plugin.slap(self, damage)

    def slay(self):
        return shinqlx.Plugin.slay(self)

    def slay_with_mod(self, mod):
        return shinqlx.slay_with_mod(self.id, mod)

    @classmethod
    def all_players(cls):
        return [
            cls(info.client_id, info=info) for info in shinqlx.players_info() if info
        ]


class AbstractDummyPlayer(Player):
    def __init__(self, name="DummyPlayer"):
        info = shinqlx.PlayerInfo(
            (
                -1,
                name,
                shinqlx.CS_CONNECTED,
                _DUMMY_USERINFO,
                -1,
                shinqlx.TEAM_SPECTATOR,
                shinqlx.PRIV_NONE,
            )
        )
        super().__init__(-1, info=info)

    @property
    def id(self):
        raise AttributeError("Dummy players do not have client IDs.")

    @property
    def steam_id(self):
        raise NotImplementedError("steam_id property needs to be implemented.")

    def update(self):
        pass

    @property
    def channel(self):
        raise NotImplementedError("channel property needs to be implemented.")

    def tell(self, msg, **kwargs):
        raise NotImplementedError("tell() needs to be implemented.")


class RconDummyPlayer(AbstractDummyPlayer):
    def __init__(self):
        super().__init__(name=self.__class__.__name__)

    @property
    def steam_id(self):
        return shinqlx.owner()

    @property
    def channel(self):
        return shinqlx.CONSOLE_CHANNEL

    def tell(self, msg, **kwargs):
        self.channel.reply(msg)



import queue
import sched
import re

import shinqlx

# ====================================================================
#                        REGULAR EXPRESSIONS
# ====================================================================

_re_say = re.compile(r"^say +\"?(?P<msg>.+)\"?$", flags=re.IGNORECASE)
_re_say_team = re.compile(r"^say_team +\"?(?P<msg>.+)\"?$", flags=re.IGNORECASE)
_re_callvote = re.compile(
    r"^(?:cv|callvote) +(?P<cmd>[^ ]+)(?: \"?(?P<args>.+?)\"?)?$", flags=re.IGNORECASE
)
_re_vote = re.compile(r"^vote +(?P<arg>.)", flags=re.IGNORECASE)
_re_team = re.compile(r"^team +(?P<arg>.)", flags=re.IGNORECASE)
_re_vote_ended = re.compile(r"^print \"Vote (?P<result>passed|failed).\n\"$")
_re_userinfo = re.compile(r"^userinfo \"(?P<vars>.+)\"$")


# ====================================================================
#                         LOW-LEVEL HANDLERS
#        These are all called by the C code, not within Python.
# ====================================================================
def handle_rcon(cmd):
    """Console commands that are to be processed as regular pyshinqlx
    commands as if the owner executes it. This allows the owner to
    interact with the Python part of shinqlx without having to connect.

    """
    # noinspection PyBroadException
    try:
        shinqlx.COMMANDS.handle_input(
            shinqlx.RconDummyPlayer(), cmd, shinqlx.CONSOLE_CHANNEL
        )
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_client_command(client_id, cmd):
    """Client commands are commands such as "say", "say_team", "scores",
    "disconnect" and so on. This function parses those and passes it
    on to the event dispatcher.

    :param: client_id: The client identifier.
    :type: client_id: int
    :param: cmd: The command being run by the client.
    :type: cmd: str

    """
    # noinspection PyBroadException
    try:
        # Dispatch the "client_command" event before further processing.
        player = shinqlx.Player(client_id)
        retval = shinqlx.EVENT_DISPATCHERS["client_command"].dispatch(player, cmd)
        if retval is False:
            return False
        if isinstance(retval, str):
            # Allow plugins to modify the command before passing it on.
            cmd = retval

        res = _re_say.match(cmd)
        if res:
            msg = res.group("msg").replace('"', "")
            channel = shinqlx.CHAT_CHANNEL
            if (
                shinqlx.EVENT_DISPATCHERS["chat"].dispatch(player, msg, channel)
                is False
            ):
                return False
            return cmd

        res = _re_say_team.match(cmd)
        if res:
            msg = res.group("msg").replace('"', "")
            if (
                player.team == "free"
            ):  # I haven't tried this, but I don't think it's even possible.
                channel = shinqlx.FREE_CHAT_CHANNEL
            elif player.team == "red":
                channel = shinqlx.RED_TEAM_CHAT_CHANNEL
            elif player.team == "blue":
                channel = shinqlx.BLUE_TEAM_CHAT_CHANNEL
            else:
                channel = shinqlx.SPECTATOR_CHAT_CHANNEL
            if (
                shinqlx.EVENT_DISPATCHERS["chat"].dispatch(player, msg, channel)
                is False
            ):
                return False
            return cmd

        res = _re_callvote.match(cmd)
        if res and not shinqlx.Plugin.is_vote_active():
            vote = res.group("cmd")
            args = res.group("args") if res.group("args") else ""
            # Set the caller for vote_started in case the vote goes through.
            # noinspection PyUnresolvedReferences
            shinqlx.EVENT_DISPATCHERS["vote_started"].caller(player)
            if (
                shinqlx.EVENT_DISPATCHERS["vote_called"].dispatch(player, vote, args)
                is False
            ):
                return False
            return cmd

        res = _re_vote.match(cmd)
        if res and shinqlx.Plugin.is_vote_active():
            arg = res.group("arg").lower()
            if (
                arg in ["y", "1"]
                and shinqlx.EVENT_DISPATCHERS["vote"].dispatch(player, True) is False
            ):
                return False
            if (
                arg in ["n", "2"]
                and shinqlx.EVENT_DISPATCHERS["vote"].dispatch(player, False) is False
            ):
                return False
            return cmd

        res = _re_team.match(cmd)
        if res:
            arg = res.group("arg").lower()
            target_team = ""
            if arg == player.team[0]:
                # Don't trigger if player is joining the same team.
                return cmd
            if arg == "f":
                target_team = "free"
            elif arg == "r":
                target_team = "red"
            elif arg == "b":
                target_team = "blue"
            elif arg == "s":
                target_team = "spectator"
            elif arg == "a":
                target_team = "any"

            if (
                target_team
                and shinqlx.EVENT_DISPATCHERS["team_switch_attempt"].dispatch(
                    player, player.team, target_team
                )
                is False
            ):
                return False
            return cmd

        res = _re_userinfo.match(cmd)
        if res:
            new_info = shinqlx.parse_variables(res.group("vars"), ordered=True)
            old_info = player.cvars
            changed = {}

            for key in new_info:
                if key not in old_info or (
                    key in old_info and new_info[key] != old_info[key]
                ):
                    changed[key] = new_info[key]

            if changed:
                ret = shinqlx.EVENT_DISPATCHERS["userinfo"].dispatch(player, changed)
                if ret is False:
                    return False
                if isinstance(ret, dict):
                    for key in ret:
                        new_info[key] = ret[key]
                    formatted_key_values = "".join(
                        [f"\\{key}\\{value}" for key, value in new_info.items()]
                    )
                    cmd = f'userinfo "{formatted_key_values}"'

        return cmd
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_server_command(client_id, cmd):
    # noinspection PyBroadException
    try:
        # Dispatch the "server_command" event before further processing.
        try:
            player = shinqlx.Player(client_id) if client_id >= 0 else None
        except shinqlx.NonexistentPlayerError:
            return True

        retval = shinqlx.EVENT_DISPATCHERS["server_command"].dispatch(player, cmd)
        if retval is False:
            return False
        if isinstance(retval, str):
            cmd = retval

        res = _re_vote_ended.match(cmd)
        if res:
            if res.group("result") == "passed":
                shinqlx.EVENT_DISPATCHERS["vote_ended"].dispatch(True)
            else:
                shinqlx.EVENT_DISPATCHERS["vote_ended"].dispatch(False)

        return cmd
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


# Executing tasks right before a frame, by the main thread, will often be desirable to avoid
# weird behavior if you were to use threading. This list will act as a task queue.
# Tasks can be added by simply adding the @shinqlx.next_frame decorator to functions.
frame_tasks = sched.scheduler()
next_frame_tasks = queue.Queue()  # type: ignore


def handle_frame():
    """This will be called every frame. To allow threads to call stuff from the
    main thread, tasks can be scheduled using the :func:`shinqlx.next_frame` decorator
    and have it be executed here.

    """

    while True:
        # This will run all tasks that are currently scheduled.
        # If one of the tasks throw an exception, it'll log it
        # and continue execution of the next tasks if any.
        # noinspection PyBroadException
        try:
            frame_tasks.run(blocking=False)
            break
        except:  # noqa: E722
            shinqlx.log_exception()
            continue
    # noinspection PyBroadException
    try:
        shinqlx.EVENT_DISPATCHERS["frame"].dispatch()
    except:  # noqa: E722
        shinqlx.log_exception()
        return True

    while not next_frame_tasks.empty():
        func, args, kwargs = next_frame_tasks.get(block=False)
        frame_tasks.enter(0, 1, func, args, kwargs)


_zmq_warning_issued = False
_first_game = True
_ad_round_number = 0


def handle_new_game(is_restart):
    # This is called early in the launch process, so it's a good place to initialize
    # shinqlx stuff that needs QLDS to be initialized.
    global _first_game
    if _first_game:
        shinqlx.late_init()
        _first_game = False

        # A good place to warn the owner if ZMQ stats are disabled.
        global _zmq_warning_issued
        stats_enabled_cvar = shinqlx.get_cvar("zmq_stats_enable")
        if (
            stats_enabled_cvar is None or not bool(int(stats_enabled_cvar))
        ) and not _zmq_warning_issued:
            logger = shinqlx.get_logger()
            logger.warning(
                "Some events will not work because ZMQ stats is not enabled. "
                'Launch the server with "zmq_stats_enable 1"'
            )
            _zmq_warning_issued = True

    shinqlx.set_map_subtitles()

    if not is_restart:
        # noinspection PyBroadException
        try:
            shinqlx.EVENT_DISPATCHERS["map"].dispatch(
                shinqlx.get_cvar("mapname"), shinqlx.get_cvar("g_factory")
            )
        except:  # noqa: E722
            shinqlx.log_exception()
            return True

    # noinspection PyBroadException
    try:
        shinqlx.EVENT_DISPATCHERS["new_game"].dispatch()
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_set_configstring(index, value):
    """Called whenever the server tries to set a configstring. Can return
    False to stop the event.

    """
    global _ad_round_number

    # noinspection PyBroadException
    try:
        res = shinqlx.EVENT_DISPATCHERS["set_configstring"].dispatch(index, value)
        if res is False:
            return False
        if isinstance(res, str):
            value = res

        # VOTES
        if index == 9 and value:
            cmd = value.split()
            vote = cmd[0] if cmd else ""
            args = " ".join(cmd[1:]) if len(cmd) > 1 else ""
            shinqlx.EVENT_DISPATCHERS["vote_started"].dispatch(vote, args)
            return
        # GAME STATE CHANGES
        if index == 0:
            old_cs = shinqlx.parse_variables(shinqlx.get_configstring(index))
            if not old_cs:
                return

            new_cs = shinqlx.parse_variables(value)
            old_state = old_cs["g_gameState"]
            new_state = new_cs["g_gameState"]
            if old_state != new_state:
                if old_state == "PRE_GAME" and new_state == "IN_PROGRESS":
                    pass
                elif old_state == "PRE_GAME" and new_state == "COUNT_DOWN":
                    _ad_round_number = 1
                    shinqlx.EVENT_DISPATCHERS["game_countdown"].dispatch()
                elif (old_state == "COUNT_DOWN" and new_state == "IN_PROGRESS") or (
                    new_state == "PRE_GAME"
                    and old_state
                    in [
                        "IN_PROGRESS",
                        "COUNT_DOWN",
                    ]
                ):
                    pass
                else:
                    logger = shinqlx.get_logger()
                    logger.warning(f"UNKNOWN GAME STATES: {old_state} - {new_state}")
        # ROUND COUNTDOWN AND START
        elif index == 661:
            cvars = shinqlx.parse_variables(value)
            if cvars:
                if "turn" in cvars:
                    # it is A&D
                    if int(cvars["state"]) == 0:
                        return
                    # round cvar appears only on round countdown
                    # and first round is 0, not 1
                    try:
                        round_number = int(cvars["round"]) * 2 + 1 + int(cvars["turn"])
                        _ad_round_number = round_number
                    except KeyError:
                        round_number = _ad_round_number
                else:
                    # it is CA
                    round_number = int(cvars["round"])

                if round_number and "time" in cvars:
                    shinqlx.EVENT_DISPATCHERS["round_countdown"].dispatch(round_number)
                    return
                if round_number:
                    shinqlx.EVENT_DISPATCHERS["round_start"].dispatch(round_number)
                    return

        return res
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_player_connect(client_id, _is_bot):
    """This will be called whenever a player tries to connect. If the dispatcher
    returns False, it will not allow the player to connect and instead show them
    a message explaining why. The default message is "You are banned from this
    server.", but it can be set with :func:`shinqlx.set_ban_message`.

    :param: client_id: The client identifier.
    :type: client_id: int
    :param: _is_bot: Whether or not the player is a bot.
    :type: _is_bot: bool

    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["player_connect"].dispatch(player)
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_player_loaded(client_id):
    """This will be called whenever a player has connected and finished loading,
    meaning it'll go off a bit later than the usual "X connected" messages.
    This will not trigger on bots.

    :param: client_id: The client identifier.
    :type: client_id: int

    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["player_loaded"].dispatch(player)
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_player_disconnect(client_id, reason):
    """This will be called whenever a player disconnects.

    :param: client_id: The client identifier.
    :type: client_id: int
    :param: reason: The reason for the disconnect
    :type: reason: str

    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["player_disconnect"].dispatch(player, reason)
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_player_spawn(client_id):
    """Called when a player spawns. Note that a spectator going in free spectate mode
    makes the client spawn, so you'll want to check for that if you only want "actual"
    spawns.

    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["player_spawn"].dispatch(player)
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_kamikaze_use(client_id):
    """This will be called whenever player uses kamikaze item.

    :param: client_id: The client identifier.
    :type: client_id: int

    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["kamikaze_use"].dispatch(player)
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_kamikaze_explode(client_id, is_used_on_demand):
    """This will be called whenever kamikaze explodes.

    :param: client_id: The client identifier.
    :type: client_id: int
    :param: is_used_on_demand: Non-zero if kamikaze is used on demand.
    :type: is_used_on_demand: int


    """
    # noinspection PyBroadException
    try:
        player = shinqlx.Player(client_id)
        return shinqlx.EVENT_DISPATCHERS["kamikaze_explode"].dispatch(
            player, bool(is_used_on_demand)
        )
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_damage(target_id, attacker_id, damage, dflags, mod):
    target_player = (
        shinqlx.Player(target_id) if target_id in range(0, 64) else target_id
    )
    attacker_player = (
        shinqlx.Player(attacker_id) if attacker_id in range(0, 64) else attacker_id
    )
    # noinspection PyBroadException
    try:
        shinqlx.EVENT_DISPATCHERS["damage"].dispatch(
            target_player, attacker_player, damage, dflags, mod
        )
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


def handle_console_print(text):
    """Called whenever the server prints something to the console and when rcon is used."""
    if not text:
        return

    # noinspection PyBroadException
    try:
        # Log console output. Removes the need to have stdout logs in addition to shinqlx.log.
        shinqlx.get_logger().debug(text.rstrip("\n"))

        res = shinqlx.EVENT_DISPATCHERS["console_print"].dispatch(text)
        if res is False:
            return False

        if _print_redirection:
            global _print_buffer
            _print_buffer += text

        if isinstance(res, str):
            return res

        return text
    except:  # noqa: E722
        shinqlx.log_exception()
        return True


_print_redirection = None
_print_buffer = ""


def redirect_print(channel):
    """Redirects print output to a channel. Useful for commands that execute console commands
    and want to redirect the output to the channel instead of letting it go to the console.

    To use it, use the return value with the "with" statement.

    .. code-block:: python
        def cmd_echo(self, player, msg, channel):
            with shinqlx.redirect_print(channel):
                shinqlx.console_command("echo {}".format(" ".join(msg)))

    """

    class PrintRedirector:
        def __init__(self, _channel):
            if not isinstance(_channel, shinqlx.AbstractChannel):
                raise ValueError(
                    "The redirection channel must be an instance of shinqlx.AbstractChannel."
                )

            self.channel = _channel

        def __enter__(self):
            global _print_redirection
            _print_redirection = self.channel

        def __exit__(self, exc_type, exc_val, exc_tb):
            global _print_redirection
            self.flush()
            _print_redirection = None

        def flush(self):
            global _print_buffer
            self.channel.reply(_print_buffer)
            _print_buffer = ""

    return PrintRedirector(channel)


def register_handlers():
    shinqlx.register_handler("rcon", handle_rcon)
    shinqlx.register_handler("client_command", handle_client_command)
    shinqlx.register_handler("server_command", handle_server_command)
    shinqlx.register_handler("new_game", handle_new_game)
    shinqlx.register_handler("set_configstring", handle_set_configstring)
    shinqlx.register_handler("console_print", handle_console_print)



from typing import overload, TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Mapping
    from datetime import timedelta
    from logging import Logger

    from redis import Redis as redisRedis, ConnectionPool
    from shinqlx import Plugin, Player

class AbstractDatabase:
    _counter: int
    plugin: Plugin

    def __init__(self, plugin: Plugin) -> None: ...
    def __del__(self) -> None: ...
    @property
    def logger(self) -> Logger: ...
    def set_permission(self, player: Player | int | str, level: int) -> None: ...
    def get_permission(self, player: Player | int | str) -> int: ...
    def has_permission(self, player: Player | int | str, level: int = ...) -> bool: ...
    def set_flag(
        self, player: Player | int | str, flag: str, value: bool = ...
    ) -> None: ...
    def clear_flag(self, player: Player | int | str, flag: str) -> None: ...
    def get_flag(
        self, player: Player | int | str, flag: str, default: bool = ...
    ) -> bool: ...
    def connect(self) -> redisRedis | None: ...
    def close(self) -> None: ...

class Redis(AbstractDatabase):
    _conn: redisRedis | None
    _pool: ConnectionPool | None
    _pass: str

    def __del__(self) -> None: ...
    def __contains__(self, key: str) -> bool: ...
    def __getitem__(self, key: str) -> str: ...
    def __setitem__(self, key: str, item: str | int) -> None: ...
    def __delitem__(self, key: str) -> None: ...
    def __getattr__(self, attr: str) -> str: ...
    @property
    def r(self) -> redisRedis: ...
    def set_permission(self, player: Player | int | str, level: int) -> None: ...
    def get_permission(self, player: Player | int | str) -> int: ...
    def has_permission(self, player: Player | int | str, level: int = ...) -> bool: ...
    def set_flag(
        self, player: Player | int | str, flag: str, value: bool = ...
    ) -> None: ...
    def get_flag(
        self, player: Player | int | str, flag: str, default: bool = False
    ) -> bool: ...
    def connect(
        self,
        host: str | None = ...,
        database: int = ...,
        unix_socket: bool = ...,
        password: str | None = ...,
    ) -> redisRedis | None: ...
    def close(self) -> None: ...
    def mset(self, *args: dict, **kwargs: str | int | float | bool) -> bool: ...
    def msetnx(self, *args: dict, **kwargs: str | int | float | bool) -> bool: ...
    @overload
    async def zadd(
        self,
        name: str,
        *args: str | int | float,
        **kwargs: int | float,
    ) -> int: ...
    @overload
    async def zadd(
        self,
        name: str,
        mapping: Mapping[str, int | float],
        nx: bool = ...,
        xx: bool = ...,
        ch: bool = ...,
        incr: bool = ...,
        gt: int | float | None = ...,
        lt: int | float | None = ...,
    ) -> int: ...
    @overload
    def zincrby(
        self, name: str, value_or_amount: str, amount_or_value: int | float = ...
    ) -> float: ...
    @overload
    def zincrby(
        self, name: str, value_or_amount: int | float, amount_or_value: str
    ) -> float: ...
    @overload
    def setex(
        self, name: str, value_or_time: str, time_or_value: int | timedelta
    ) -> bool: ...
    @overload
    def setex(
        self, name: str, value_or_time: int | timedelta, time_or_value: str
    ) -> bool: ...
    @overload
    def lrem(self, name: str, value_or_count: str, num_or_value: int = ...) -> int: ...
    @overload
    def lrem(self, name: str, value_or_count: int, num_or_value: str) -> int: ...



# pylint: disable=C0413

import re as _re
import importlib

try:
    _shinqlx = importlib.import_module(name="_shinqlx")
    from _shinqlx import (
        DEBUG,
        RET_NONE,
        RET_STOP,
        RET_STOP_EVENT,
        RET_STOP_ALL,
        RET_USAGE,
        PRI_HIGHEST,
        PRI_HIGH,
        PRI_NORMAL,
        PRI_LOW,
        PRI_LOWEST,
        CVAR_ARCHIVE,
        CVAR_USERINFO,
        CVAR_SERVERINFO,
        CVAR_SYSTEMINFO,
        CVAR_INIT,
        CVAR_LATCH,
        CVAR_ROM,
        CVAR_USER_CREATED,
        CVAR_TEMP,
        CVAR_CHEAT,
        CVAR_NORESTART,
        PRIV_NONE,
        PRIV_MOD,
        PRIV_ADMIN,
        PRIV_ROOT,
        PRIV_BANNED,
        CS_FREE,
        CS_ZOMBIE,
        CS_CONNECTED,
        CS_PRIMED,
        CS_ACTIVE,
        TEAM_FREE,
        TEAM_RED,
        TEAM_BLUE,
        TEAM_SPECTATOR,
        MOD_UNKNOWN,
        MOD_SHOTGUN,
        MOD_GAUNTLET,
        MOD_MACHINEGUN,
        MOD_GRENADE,
        MOD_GRENADE_SPLASH,
        MOD_ROCKET,
        MOD_ROCKET_SPLASH,
        MOD_PLASMA,
        MOD_PLASMA_SPLASH,
        MOD_RAILGUN,
        MOD_LIGHTNING,
        MOD_BFG,
        MOD_BFG_SPLASH,
        MOD_WATER,
        MOD_SLIME,
        MOD_LAVA,
        MOD_CRUSH,
        MOD_TELEFRAG,
        MOD_FALLING,
        MOD_SUICIDE,
        MOD_TARGET_LASER,
        MOD_TRIGGER_HURT,
        MOD_NAIL,
        MOD_CHAINGUN,
        MOD_PROXIMITY_MINE,
        MOD_KAMIKAZE,
        MOD_JUICED,
        MOD_GRAPPLE,
        MOD_SWITCH_TEAMS,
        MOD_THAW,
        MOD_LIGHTNING_DISCHARGE,
        MOD_HMG,
        MOD_RAILGUN_HEADSHOT,
        DAMAGE_RADIUS,
        DAMAGE_NO_ARMOR,
        DAMAGE_NO_KNOCKBACK,
        DAMAGE_NO_PROTECTION,
        DAMAGE_NO_TEAM_PROTECTION,
        Vector3,
        Flight,
        Powerups,
        Weapons,
        PlayerInfo,
        PlayerState,
        PlayerStats,
        player_info,
        players_info,
        get_userinfo,
        send_server_command,
        client_command,
        console_command,
        get_cvar,
        set_cvar,
        set_cvar_limit,
        kick,
        console_print,
        get_configstring,
        set_configstring,
        force_vote,
        add_console_command,
        player_state,
        player_stats,
        set_position,
        set_velocity,
        noclip,
        set_health,
        set_armor,
        set_weapons,
        set_weapon,
        set_ammo,
        set_powerups,
        set_holdable,
        drop_holdable,
        set_flight,
        set_invulnerability,
        set_score,
        callvote,
        allow_single_player,
        player_spawn,
        set_privileges,
        destroy_kamikaze_timers,
        spawn_item,
        remove_dropped_items,
        slay_with_mod,
        replace_items,
        dev_print_items,
        force_weapon_respawn_time,
        register_handler,
        get_targetting_entities,
    )
except ModuleNotFoundError:
    _shinqlx = importlib.import_module(name="._shinqlx", package="shinqlx")
    from ._shinqlx import (
        DEBUG,
        RET_NONE,
        RET_STOP,
        RET_STOP_EVENT,
        RET_STOP_ALL,
        RET_USAGE,
        PRI_HIGHEST,
        PRI_HIGH,
        PRI_NORMAL,
        PRI_LOW,
        PRI_LOWEST,
        CVAR_ARCHIVE,
        CVAR_USERINFO,
        CVAR_SERVERINFO,
        CVAR_SYSTEMINFO,
        CVAR_INIT,
        CVAR_LATCH,
        CVAR_ROM,
        CVAR_USER_CREATED,
        CVAR_TEMP,
        CVAR_CHEAT,
        CVAR_NORESTART,
        PRIV_NONE,
        PRIV_MOD,
        PRIV_ADMIN,
        PRIV_ROOT,
        PRIV_BANNED,
        CS_FREE,
        CS_ZOMBIE,
        CS_CONNECTED,
        CS_PRIMED,
        CS_ACTIVE,
        TEAM_FREE,
        TEAM_RED,
        TEAM_BLUE,
        TEAM_SPECTATOR,
        MOD_UNKNOWN,
        MOD_SHOTGUN,
        MOD_GAUNTLET,
        MOD_MACHINEGUN,
        MOD_GRENADE,
        MOD_GRENADE_SPLASH,
        MOD_ROCKET,
        MOD_ROCKET_SPLASH,
        MOD_PLASMA,
        MOD_PLASMA_SPLASH,
        MOD_RAILGUN,
        MOD_LIGHTNING,
        MOD_BFG,
        MOD_BFG_SPLASH,
        MOD_WATER,
        MOD_SLIME,
        MOD_LAVA,
        MOD_CRUSH,
        MOD_TELEFRAG,
        MOD_FALLING,
        MOD_SUICIDE,
        MOD_TARGET_LASER,
        MOD_TRIGGER_HURT,
        MOD_NAIL,
        MOD_CHAINGUN,
        MOD_PROXIMITY_MINE,
        MOD_KAMIKAZE,
        MOD_JUICED,
        MOD_GRAPPLE,
        MOD_SWITCH_TEAMS,
        MOD_THAW,
        MOD_LIGHTNING_DISCHARGE,
        MOD_HMG,
        MOD_RAILGUN_HEADSHOT,
        Vector3,
        Flight,
        Powerups,
        Weapons,
        PlayerInfo,
        PlayerState,
        PlayerStats,
        player_info,
        players_info,
        get_userinfo,
        send_server_command,
        client_command,
        console_command,
        get_cvar,
        set_cvar,
        set_cvar_limit,
        kick,
        console_print,
        get_configstring,
        set_configstring,
        force_vote,
        add_console_command,
        player_state,
        player_stats,
        set_position,
        set_velocity,
        noclip,
        set_health,
        set_armor,
        set_weapons,
        set_weapon,
        set_ammo,
        set_powerups,
        set_holdable,
        drop_holdable,
        set_flight,
        set_invulnerability,
        set_score,
        callvote,
        allow_single_player,
        player_spawn,
        set_privileges,
        destroy_kamikaze_timers,
        spawn_item,
        remove_dropped_items,
        slay_with_mod,
        replace_items,
        dev_print_items,
        force_weapon_respawn_time,
        register_handler,
        get_targetting_entities,
    )

from ._core import (
    PluginLoadError,
    PluginUnloadError,
    TEAMS,
    GAMETYPES,
    GAMETYPES_SHORT,
    CONNECTION_STATES,
    WEAPONS,
    DEFAULT_PLUGINS,
    parse_variables,
    get_logger,
    log_exception,
    handle_exception,
    threading_excepthook,
    uptime,
    owner,
    stats_listener,
    set_cvar_once,
    set_cvar_limit_once,
    set_plugins_version,
    set_map_subtitles,
    next_frame,
    delay,
    thread,
    load_preset_plugins,
    load_plugin,
    unload_plugin,
    reload_plugin,
    initialize_cvars,
    initialize,
    late_init,
)
from ._game import Game, NonexistentGameError
from ._player import (
    Player,
    NonexistentPlayerError,
    AbstractDummyPlayer,
    RconDummyPlayer,
)
from ._plugin import Plugin
from ._events import (
    EventDispatcher,
    EventDispatcherManager,
    ConsolePrintDispatcher,
    CommandDispatcher,
    ClientCommandDispatcher,
    ServerCommandDispatcher,
    FrameEventDispatcher,
    SetConfigstringDispatcher,
    ChatEventDispatcher,
    UnloadDispatcher,
    PlayerConnectDispatcher,
    PlayerLoadedDispatcher,
    PlayerDisonnectDispatcher,
    PlayerSpawnDispatcher,
    StatsDispatcher,
    VoteCalledDispatcher,
    VoteStartedDispatcher,
    VoteEndedDispatcher,
    VoteDispatcher,
    GameCountdownDispatcher,
    GameStartDispatcher,
    GameEndDispatcher,
    RoundCountdownDispatcher,
    RoundStartDispatcher,
    RoundEndDispatcher,
    TeamSwitchDispatcher,
    TeamSwitchAttemptDispatcher,
    MapDispatcher,
    NewGameDispatcher,
    KillDispatcher,
    DeathDispatcher,
    UserinfoDispatcher,
    KamikazeUseDispatcher,
    KamikazeExplodeDispatcher,
    PlayerItemsTossDispatcher,
    DamageDispatcher,
    EVENT_DISPATCHERS,
)
from ._handlers import (
    frame_tasks,
    next_frame_tasks,
    handle_rcon,
    handle_client_command,
    handle_server_command,
    handle_frame,
    handle_new_game,
    handle_set_configstring,
    handle_player_connect,
    handle_player_loaded,
    handle_player_disconnect,
    handle_player_spawn,
    handle_kamikaze_use,
    handle_kamikaze_explode,
    handle_console_print,
    handle_damage,
    redirect_print,
    register_handlers,
)
from ._commands import (
    MAX_MSG_LENGTH,
    re_color_tag,
    AbstractChannel,
    ChatChannel,
    TeamChatChannel,
    TellChannel,
    ConsoleChannel,
    ClientCommandChannel,
    Command,
    CommandInvoker,
    COMMANDS,
    CHAT_CHANNEL,
    RED_TEAM_CHAT_CHANNEL,
    BLUE_TEAM_CHAT_CHANNEL,
    FREE_CHAT_CHANNEL,
    SPECTATOR_CHAT_CHANNEL,
    CONSOLE_CHANNEL,
)
from ._zmq import StatsListener

__version__ = _shinqlx.__version__
__plugins_version__ = "NOT_SET"

_map_title = ""
_map_subtitle1 = ""
_map_subtitle2 = ""

temp = _re.search(r"(\d+)\.(\d+)\.(\d+)", __version__)
if temp is None:
    __version_info__ = (999, 999, 999)
else:
    # noinspection PyBroadException
    try:
        __version_info__ = int(temp.group(1)), int(temp.group(2)), int(temp.group(3))
    except:  # noqa: E722
        __version_info__ = (999, 999, 999)
del temp

# Put everything into a single module.
__all__ = [
    "__version__",
    "__version_info__",
    "__plugins_version__",
    # _shinqlx
    "DEBUG",
    "RET_NONE",
    "RET_STOP",
    "RET_STOP_EVENT",
    "RET_STOP_ALL",
    "RET_USAGE",
    "PRI_HIGHEST",
    "PRI_HIGH",
    "PRI_NORMAL",
    "PRI_LOW",
    "PRI_LOWEST",
    "CVAR_ARCHIVE",
    "CVAR_USERINFO",
    "CVAR_SERVERINFO",
    "CVAR_SYSTEMINFO",
    "CVAR_INIT",
    "CVAR_LATCH",
    "CVAR_ROM",
    "CVAR_USER_CREATED",
    "CVAR_TEMP",
    "CVAR_CHEAT",
    "CVAR_NORESTART",
    "PRIV_NONE",
    "PRIV_MOD",
    "PRIV_ADMIN",
    "PRIV_ROOT",
    "PRIV_BANNED",
    "CS_FREE",
    "CS_ZOMBIE",
    "CS_CONNECTED",
    "CS_PRIMED",
    "CS_ACTIVE",
    "TEAM_FREE",
    "TEAM_RED",
    "TEAM_BLUE",
    "TEAM_SPECTATOR",
    "MOD_UNKNOWN",
    "MOD_SHOTGUN",
    "MOD_GAUNTLET",
    "MOD_MACHINEGUN",
    "MOD_GRENADE",
    "MOD_GRENADE_SPLASH",
    "MOD_ROCKET",
    "MOD_ROCKET_SPLASH",
    "MOD_PLASMA",
    "MOD_PLASMA_SPLASH",
    "MOD_RAILGUN",
    "MOD_LIGHTNING",
    "MOD_BFG",
    "MOD_BFG_SPLASH",
    "MOD_WATER",
    "MOD_SLIME",
    "MOD_LAVA",
    "MOD_CRUSH",
    "MOD_TELEFRAG",
    "MOD_FALLING",
    "MOD_SUICIDE",
    "MOD_TARGET_LASER",
    "MOD_TRIGGER_HURT",
    "MOD_NAIL",
    "MOD_CHAINGUN",
    "MOD_PROXIMITY_MINE",
    "MOD_KAMIKAZE",
    "MOD_JUICED",
    "MOD_GRAPPLE",
    "MOD_SWITCH_TEAMS",
    "MOD_THAW",
    "MOD_LIGHTNING_DISCHARGE",
    "MOD_HMG",
    "MOD_RAILGUN_HEADSHOT",
    "DAMAGE_RADIUS",
    "DAMAGE_NO_ARMOR",
    "DAMAGE_NO_KNOCKBACK",
    "DAMAGE_NO_PROTECTION",
    "DAMAGE_NO_TEAM_PROTECTION",
    "Vector3",
    "Flight",
    "Powerups",
    "Weapons",
    "PlayerInfo",
    "PlayerState",
    "PlayerStats",
    "player_info",
    "players_info",
    "get_userinfo",
    "send_server_command",
    "client_command",
    "console_command",
    "get_cvar",
    "set_cvar",
    "set_cvar_limit",
    "kick",
    "console_print",
    "get_configstring",
    "set_configstring",
    "force_vote",
    "add_console_command",
    "player_state",
    "player_stats",
    "set_position",
    "set_velocity",
    "noclip",
    "set_health",
    "set_armor",
    "set_weapons",
    "set_weapon",
    "set_ammo",
    "set_powerups",
    "set_holdable",
    "drop_holdable",
    "set_flight",
    "set_invulnerability",
    "set_score",
    "callvote",
    "allow_single_player",
    "player_spawn",
    "set_privileges",
    "destroy_kamikaze_timers",
    "spawn_item",
    "remove_dropped_items",
    "slay_with_mod",
    "replace_items",
    "dev_print_items",
    "force_weapon_respawn_time",
    "register_handler",
    "get_targetting_entities",
    # _core
    "PluginLoadError",
    "PluginUnloadError",
    "TEAMS",
    "GAMETYPES",
    "GAMETYPES_SHORT",
    "CONNECTION_STATES",
    "WEAPONS",
    "DEFAULT_PLUGINS",
    "parse_variables",
    "get_logger",
    "log_exception",
    "handle_exception",
    "threading_excepthook",
    "uptime",
    "owner",
    "stats_listener",
    "set_cvar_once",
    "set_cvar_limit_once",
    "set_plugins_version",
    "set_map_subtitles",
    "next_frame",
    "delay",
    "thread",
    "load_preset_plugins",
    "load_plugin",
    "unload_plugin",
    "reload_plugin",
    "initialize_cvars",
    "initialize",
    "late_init",
    # _plugin, _game
    "Plugin",
    "Game",
    "NonexistentGameError",
    # _player
    "Player",
    "NonexistentPlayerError",
    "AbstractDummyPlayer",
    "RconDummyPlayer",
    # _commands
    "MAX_MSG_LENGTH",
    "re_color_tag",
    "AbstractChannel",
    "ChatChannel",
    "TeamChatChannel",
    "TellChannel",
    "ConsoleChannel",
    "ClientCommandChannel",
    "Command",
    "CommandInvoker",
    "COMMANDS",
    "CHAT_CHANNEL",
    "RED_TEAM_CHAT_CHANNEL",
    "BLUE_TEAM_CHAT_CHANNEL",
    "FREE_CHAT_CHANNEL",
    "SPECTATOR_CHAT_CHANNEL",
    "CONSOLE_CHANNEL",
    # _events
    "EventDispatcher",
    "EventDispatcherManager",
    "ConsolePrintDispatcher",
    "CommandDispatcher",
    "ClientCommandDispatcher",
    "ServerCommandDispatcher",
    "FrameEventDispatcher",
    "SetConfigstringDispatcher",
    "ChatEventDispatcher",
    "UnloadDispatcher",
    "PlayerConnectDispatcher",
    "PlayerLoadedDispatcher",
    "PlayerDisonnectDispatcher",
    "PlayerSpawnDispatcher",
    "StatsDispatcher",
    "VoteCalledDispatcher",
    "VoteStartedDispatcher",
    "VoteEndedDispatcher",
    "VoteDispatcher",
    "GameCountdownDispatcher",
    "GameStartDispatcher",
    "GameEndDispatcher",
    "RoundCountdownDispatcher",
    "RoundStartDispatcher",
    "RoundEndDispatcher",
    "TeamSwitchDispatcher",
    "TeamSwitchAttemptDispatcher",
    "MapDispatcher",
    "NewGameDispatcher",
    "KillDispatcher",
    "DeathDispatcher",
    "UserinfoDispatcher",
    "KamikazeUseDispatcher",
    "KamikazeExplodeDispatcher",
    "PlayerItemsTossDispatcher",
    "DamageDispatcher",
    "EVENT_DISPATCHERS",
    # _handlers
    "frame_tasks",
    "next_frame_tasks",
    "handle_rcon",
    "handle_client_command",
    "handle_server_command",
    "handle_frame",
    "handle_new_game",
    "handle_set_configstring",
    "handle_player_connect",
    "handle_player_loaded",
    "handle_player_disconnect",
    "handle_player_spawn",
    "handle_kamikaze_use",
    "handle_kamikaze_explode",
    "handle_console_print",
    "handle_damage",
    "redirect_print",
    "register_handlers",
    # _zmq
    "StatsListener",
]



import re

import shinqlx

_re_vote = re.compile(r"^(?P<cmd>[^ ]+)(?: \"?(?P<args>.*?)\"?)?$")
hot_plugged_events = (
    "frame",
    "player_connect",
    "player_loaded",
    "player_disconnect",
    "player_spawn",
    "kamikaze_use",
    "kamikaze_explode",
    "damage",
)


# ====================================================================
#                               EVENTS
# ====================================================================
class EventDispatcher:
    """The base event dispatcher. Each event should inherit this and provides a way
    to hook into events by registering an event handler.

    """

    no_debug = (
        "frame",
        "set_configstring",
        "stats",
        "server_command",
        "death",
        "kill",
        "command",
        "console_print",
        "damage",
    )
    need_zmq_stats_enabled = False

    def __init__(self):
        self.name: str
        self.plugins = {}
        self._args = None
        self._kwargs = None
        self._return_value = True

    @property
    def args(self):
        return self._args

    # noinspection PyUnresolvedReferences
    @args.setter
    def args(self, value):
        self._args = value

    @property
    def kwargs(self):
        return self._kwargs

    # noinspection PyUnresolvedReferences
    @kwargs.setter
    def kwargs(self, value):
        self._kwargs = value

    @property
    def return_value(self):
        return self._return_value

    # noinspection PyUnresolvedReferences
    @return_value.setter
    def return_value(self, value):
        self._return_value = value

    def dispatch(self, *args, **kwargs):
        """Calls all the handlers that have been registered when hooking this event.
        The recommended way to use this for events that inherit this class is to
        override the method with explicit arguments (as opposed to this one's)
        and call this method by using ``super().dispatch()``.

        Handlers have several options for return values that can affect the flow:
            - shinqlx.RET_NONE or None -- Continue execution normally.
            - shinqlx.RET_STOP -- Stop any further handlers from being called.
            - shinqlx.RET_STOP_EVENT -- Let handlers process it, but stop the event
                at the engine-level.
            - shinqlx.RET_STOP_ALL -- Stop handlers **and** the event.
            - Any other value -- Passed on to :func:`self.handle_return`, which will
                by default simply send a warning to the logger about an unknown value
                being returned. Can be overridden so that events can have their own
                special return values.

        :param: args: Any arguments.
        :param: kwargs: Any keyword arguments.

        """
        # Allow subclasses of this to edit the arguments without having
        # to reimplement this method. Whenever an unknown return value
        # is returned, we pass it on to handle_return.
        self.args = args
        self.kwargs = kwargs
        logger = shinqlx.get_logger()
        # Log the events as they come in.
        if self.name not in self.no_debug:
            dbgstr = f"{self.name}{args}"
            if len(dbgstr) > 100:
                dbgstr = dbgstr[0:99] + ")"
            logger.debug(dbgstr)

        plugins = self.plugins.copy()
        self.return_value = True
        for i in range(5):
            for plugin_name, plugin in plugins.items():
                for handler in plugin[i]:
                    # noinspection PyBroadException
                    try:
                        res = handler(*self.args, **self.kwargs)
                        if res == shinqlx.RET_NONE or res is None:
                            continue
                        if res == shinqlx.RET_STOP:
                            return True
                        if res == shinqlx.RET_STOP_EVENT:
                            self.return_value = False
                            continue
                        if res == shinqlx.RET_STOP_ALL:
                            return False
                        # Got an unknown return value.
                        return_handler = self.handle_return(handler, res)
                        if return_handler is not None:
                            return return_handler
                    except:  # noqa: E722
                        shinqlx.log_exception(plugin_name)
                        continue

        return self.return_value

    def handle_return(self, handler, value):
        """Handle an unknown return value. If this returns anything but None,
        it will stop execution of the event and pass the return value on
        to the C-level handlers. This method can be useful to override,
        because of the fact that you can edit the arguments that will be
        passed on to any handler after whatever handler returned *value*
        by editing *self.args*, *self.kwargs*. Furthermore, *self.return_value*
        is the return value that will be sent to the C-level handler if the
        event isn't stopped later along the road.
        """
        logger = shinqlx.get_logger()
        logger.warning(
            "Handler '%s' returned unknown value '%s' for event '%s'",
            handler.__name__,
            value,
            self.name,
        )

    def add_hook(self, plugin, handler, priority=shinqlx.PRI_NORMAL):
        """Hook the event, making the handler get called with relevant arguments
        whenever the event is takes place.

        :param: plugin: The plugin that's hooking the event.
        :type: plugin: shinqlx.Plugin
        :param: handler: The handler to be called when the event takes place.
        :type: handler: callable
        :param: priority: The priority of the hook. Determines the order the handlers are called in.
        :type: priority: shinqlx.PRI_LOWEST, shinqlx.PRI_LOW, shinqlx.PRI_NORMAL, shinqlx.PRI_HIGH or
            shinqlx.PRI_HIGHEST
        :raises: ValueError

        """
        if not shinqlx.PRI_HIGHEST <= priority <= shinqlx.PRI_LOWEST:
            raise ValueError(f"'{priority}' is an invalid priority level.")

        stats_enable_cvar = shinqlx.get_cvar("zmq_stats_enable")
        if self.need_zmq_stats_enabled and (
            stats_enable_cvar is None or not bool(int(stats_enable_cvar))
        ):
            raise AssertionError(
                f"{self.name} hook requires zmq_stats_enabled cvar to have nonzero value"
            )

        if self.name in hot_plugged_events and len(self.plugins) == 0:
            shinqlx.register_handler(self.name, getattr(shinqlx, f"handle_{self.name}"))

        if plugin not in self.plugins:
            # Initialize tuple.
            self.plugins[plugin] = ([], [], [], [], [])  # 5 priority levels.
        else:
            # Check if we've already registered this handler.
            for i in range(len(self.plugins[plugin])):
                for hook in self.plugins[plugin][i]:
                    if handler == hook:
                        raise ValueError(
                            "The event has already been hooked with the same handler and priority."
                        )

        self.plugins[plugin][priority].append(handler)

    def remove_hook(self, plugin, handler, priority=shinqlx.PRI_NORMAL):
        """Removes a previously hooked event.

        :param: plugin: The plugin that hooked the event.
        :type: plugin: shinqlx.Plugin
        :param: handler: The handler used when hooked.
        :type: handler: callable
        :param: priority: The priority of the hook when hooked.
        :type: priority: shinqlx.PRI_LOWEST, shinqlx.PRI_LOW, shinqlx.PRI_NORMAL, shinqlx.PRI_HIGH or
            shinqlx.PRI_HIGHEST
        :raises: ValueError

        """
        for hook in self.plugins[plugin][priority]:
            if handler == hook:
                self.plugins[plugin][priority].remove(handler)
                if self.name in hot_plugged_events and len(self.plugins) == 0:
                    shinqlx.register_handler(self.name, None)

                return

        raise ValueError("The event has not been hooked with the handler provided")


class EventDispatcherManager:
    """Holds all the event dispatchers and provides a way to access the dispatcher
    instances by accessing it like a dictionary using the event name as a key.
    Only one dispatcher can be used per event.

    """

    def __init__(self):
        self._dispatchers = {}

    def __getitem__(self, key):
        return self._dispatchers[key]

    def __contains__(self, key):
        return key in self._dispatchers

    def add_dispatcher(self, dispatcher):
        if dispatcher.name in self:
            raise ValueError("Event name already taken.")
        if not issubclass(dispatcher, EventDispatcher):
            raise ValueError(
                "Cannot add an event dispatcher not based on EventDispatcher."
            )

        self._dispatchers[dispatcher.name] = dispatcher()

    def remove_dispatcher(self, dispatcher) -> None:
        if dispatcher.name not in self:
            raise ValueError("Event name not found.")

        del self._dispatchers[dispatcher.name]

    def remove_dispatcher_by_name(self, event_name) -> None:
        if event_name not in self:
            raise ValueError("Event name not found.")

        del self._dispatchers[event_name]


# ====================================================================
#                          EVENT DISPATCHERS
# ====================================================================
class ConsolePrintDispatcher(EventDispatcher):
    """Event that goes off whenever the console prints something, including
    those with :func:`shinqlx.console_print`.

    """

    name = "console_print"

    def dispatch(self, text):
        return super().dispatch(text)

    def handle_return(self, handler, value):
        """If a string was returned, continue execution, but we edit the
        string that's being printed along the chain of handlers.

        """
        if isinstance(value, str):
            self.args = (value,)
            self.return_value = value
        else:
            return super().handle_return(handler, value)


class CommandDispatcher(EventDispatcher):
    """Event that goes off when a command is executed. This can be used
    to for instance keep a log of all the commands admins have used.

    """

    name = "command"

    def dispatch(self, caller, command, args):
        super().dispatch(caller, command, args)


class ClientCommandDispatcher(EventDispatcher):
    """Event that triggers with any client command. This overlaps with
    other events, such as "chat".

    """

    name = "client_command"

    def dispatch(self, player, cmd):
        ret = super().dispatch(player, cmd)
        if ret is False:
            return False

        ret = shinqlx.COMMANDS.handle_input(
            player, cmd, shinqlx.ClientCommandChannel(player)
        )
        if ret is False:
            return False

        return self.return_value

    def handle_return(self, handler, value):
        """If a string was returned, continue execution, but we edit the
        command that's being executed along the chain of handlers.

        """
        if isinstance(value, str):
            player, cmd = self.args
            self.args = (player, value)
            self.return_value = value
        else:
            return super().handle_return(handler, value)


class ServerCommandDispatcher(EventDispatcher):
    """Event that triggers with any server command sent by the server,
    including :func:`shinqlx.send_server_command`. Can be cancelled.

    """

    name = "server_command"

    def dispatch(self, player, cmd):
        return super().dispatch(player, cmd)

    def handle_return(self, handler, value):
        """If a string was returned, continue execution, but we edit the
        command that's being sent along the chain of handlers.

        """
        if isinstance(value, str):
            player, cmd = self.args
            self.args = (player, value)
            self.return_value = value
        else:
            return super().handle_return(handler, value)


class FrameEventDispatcher(EventDispatcher):
    """Event that triggers every frame. Cannot be cancelled."""

    name = "frame"

    def dispatch(self):
        return super().dispatch()


class SetConfigstringDispatcher(EventDispatcher):
    """Event that triggers when the server tries to set a configstring. You can
    stop this event and use :func:`shinqlx.set_configstring` to modify it, but a
    more elegant way to do it is simply returning the new configstring in
    the handler, and the modified one will go down the plugin chain instead.

    """

    name = "set_configstring"

    def dispatch(self, index, value):
        return super().dispatch(index, value)

    def handle_return(self, handler, value):
        """If a string was returned, continue execution, but we edit the
        configstring to the returned string. This allows multiple handlers
        to edit the configstring along the way before it's actually
        set by the QL engine.

        """
        if isinstance(value, str):
            index, old_value = self.args
            self.args = (index, value)
            self.return_value = value
        else:
            return super().handle_return(handler, value)


class ChatEventDispatcher(EventDispatcher):
    """Event that triggers with the "say" command. If the handler cancels it,
    the message will also be cancelled.

    """

    name = "chat"

    def dispatch(self, player, msg, channel):
        ret = shinqlx.COMMANDS.handle_input(player, msg, channel)
        if ret is False:  # Stop event if told to.
            return False

        return super().dispatch(player, msg, channel)


class UnloadDispatcher(EventDispatcher):
    """Event that triggers whenever a plugin is unloaded. Cannot be cancelled."""

    name = "unload"

    def dispatch(self, plugin):
        super().dispatch(plugin)


class PlayerConnectDispatcher(EventDispatcher):
    """Event that triggers whenever a player tries to connect. If the event
    is not stopped, it will let the player connect as usual. If it is stopped
    it will either display a generic ban message, or whatever string is returned
    by the handler.

    """

    name = "player_connect"

    def dispatch(self, player):
        return super().dispatch(player)

    def handle_return(self, handler, value):
        """If a string was returned, stop execution of event, disallow
        the player from connecting, and display the returned string as
        a message to the player trying to connect.

        """
        if isinstance(value, str):
            return value
        return super().handle_return(handler, value)


class PlayerLoadedDispatcher(EventDispatcher):
    """Event that triggers whenever a player connects *and* finishes loading.
    This means it'll trigger later than the "X connected" messages in-game,
    and it will also trigger when a map changes and players finish loading it.

    """

    name = "player_loaded"

    def dispatch(self, player):
        return super().dispatch(player)


class PlayerDisonnectDispatcher(EventDispatcher):
    """Event that triggers whenever a player disconnects. Cannot be cancelled."""

    name = "player_disconnect"

    def dispatch(self, player, reason):
        return super().dispatch(player, reason)


class PlayerSpawnDispatcher(EventDispatcher):
    """Event that triggers when a player spawns. Cannot be cancelled."""

    name = "player_spawn"

    def dispatch(self, player):
        return super().dispatch(player)


class StatsDispatcher(EventDispatcher):
    """Event that triggers whenever the server sends stats over ZMQ."""

    name = "stats"
    need_zmq_stats_enabled = True

    def dispatch(self, stats):
        return super().dispatch(stats)


class VoteCalledDispatcher(EventDispatcher):
    """Event that goes off whenever a player tries to call a vote. Note that
    this goes off even if it's a vote command that is invalid. Use vote_started
    if you only need votes that actually go through. Use this one for custom votes.

    """

    name = "vote_called"

    def dispatch(self, player, vote, args):
        return super().dispatch(player, vote, args)


class VoteStartedDispatcher(EventDispatcher):
    """Event that goes off whenever a vote starts. A vote started with Plugin.callvote()
    will have the caller set to None.

    """

    name = "vote_started"

    def __init__(self):
        super().__init__()
        self._caller = None

    def dispatch(self, vote, args):
        return super().dispatch(self._caller, vote, args)

    def caller(self, player):
        self._caller = player


class VoteEndedDispatcher(EventDispatcher):
    """Event that goes off whenever a vote either passes or fails."""

    name = "vote_ended"

    def dispatch(self, passed):
        # Check if there's a current vote in the first place.
        cs = shinqlx.get_configstring(9)
        if not cs:
            shinqlx.get_logger().warning("vote_ended went off without configstring 9.")
            return

        res = _re_vote.match(cs)
        if not res:
            shinqlx.get_logger().warning(f"invalid vote called: {cs}")
            return

        vote = res.group("cmd")
        args = res.group("args") if res.group("args") else ""
        votes = (int(shinqlx.get_configstring(10)), int(shinqlx.get_configstring(11)))
        super().dispatch(votes, vote, args, passed)


class VoteDispatcher(EventDispatcher):
    """Event that goes off whenever someone tries to vote either yes or no."""

    name = "vote"

    def dispatch(self, player, yes):
        return super().dispatch(player, yes)


class GameCountdownDispatcher(EventDispatcher):
    """Event that goes off when the countdown before a game starts."""

    name = "game_countdown"

    def dispatch(self):
        return super().dispatch()


class GameStartDispatcher(EventDispatcher):
    """Event that goes off when a game starts."""

    name = "game_start"
    need_zmq_stats_enabled = True

    def dispatch(self, data):
        return super().dispatch(data)


class GameEndDispatcher(EventDispatcher):
    """Event that goes off when a game ends."""

    name = "game_end"
    need_zmq_stats_enabled = True

    def dispatch(self, data):
        return super().dispatch(data)


class RoundCountdownDispatcher(EventDispatcher):
    """Event that goes off when the countdown before a round starts."""

    name = "round_countdown"

    def dispatch(self, round_number):
        return super().dispatch(round_number)


class RoundStartDispatcher(EventDispatcher):
    """Event that goes off when a round starts."""

    name = "round_start"

    def dispatch(self, round_number):
        return super().dispatch(round_number)


class RoundEndDispatcher(EventDispatcher):
    """Event that goes off when a round ends."""

    name = "round_end"
    need_zmq_stats_enabled = True

    def dispatch(self, data):
        return super().dispatch(data)


class TeamSwitchDispatcher(EventDispatcher):
    """For when a player switches teams. If cancelled,
    simply put the player back in the old team.

    If possible, consider using team_switch_attempt for a cleaner
    solution if you need to cancel the event."""

    name = "team_switch"
    need_zmq_stats_enabled = True

    def dispatch(self, player, old_team, new_team):
        return super().dispatch(player, old_team, new_team)


class TeamSwitchAttemptDispatcher(EventDispatcher):
    """For when a player attempts to join a team. Prevents the player from doing it when cancelled.

    When players click the Join Match button, it sends "team a" (with the "a" being "any",
    presumably), meaning the new_team argument can also be "any" in addition to all the
    other teams.

    """

    name = "team_switch_attempt"

    def dispatch(self, player, old_team, new_team):
        return super().dispatch(player, old_team, new_team)


class MapDispatcher(EventDispatcher):
    """Event that goes off when a map is loaded, even if the same map is loaded again."""

    name = "map"

    def dispatch(self, mapname, factory):
        return super().dispatch(mapname, factory)


class NewGameDispatcher(EventDispatcher):
    """Event that goes off when the game module is initialized. This happens when new maps are loaded,
    a game is aborted, a game ends but stays on the same map, or when the game itself starts.

    """

    name = "new_game"

    def dispatch(self):
        return super().dispatch()


class KillDispatcher(EventDispatcher):
    """Event that goes off when someone is killed."""

    name = "kill"
    need_zmq_stats_enabled = True

    def dispatch(self, victim, killer, data):
        return super().dispatch(victim, killer, data)


class DeathDispatcher(EventDispatcher):
    """Event that goes off when someone dies."""

    name = "death"
    need_zmq_stats_enabled = True

    def dispatch(self, victim, killer, data):
        return super().dispatch(victim, killer, data)


class UserinfoDispatcher(EventDispatcher):
    """Event for clients changing their userinfo."""

    name = "userinfo"

    def dispatch(self, player, changed):
        return super().dispatch(player, changed)

    def handle_return(self, handler, value):
        """Takes a returned dictionary and applies it to the current userinfo."""
        if isinstance(value, dict):
            player, changed = self.args
            self.args = (player, changed)
            self.return_value = changed
        else:
            return super().handle_return(handler, value)


class KamikazeUseDispatcher(EventDispatcher):
    """Event that goes off when player uses kamikaze item."""

    name = "kamikaze_use"

    def dispatch(self, player):
        return super().dispatch(player)


class KamikazeExplodeDispatcher(EventDispatcher):
    """Event that goes off when kamikaze explodes."""

    name = "kamikaze_explode"

    def dispatch(self, player, is_used_on_demand):
        return super().dispatch(player, is_used_on_demand)


class PlayerItemsTossDispatcher(EventDispatcher):
    """Event that goes off when player's items are dropeed (on death or disconnects)."""

    name = "player_items_toss"

    def dispatch(self, player):
        return super().dispatch(player)

    def handle_return(self, handler, value):
        if isinstance(value, (int, list, str)):
            self.return_value = value
        else:
            return super().handle_return(handler, value)


class DamageDispatcher(EventDispatcher):
    """Event that goes off when someone is inflicted with damage."""

    name = "damage"
    need_zmq_stats_enabled = False

    def dispatch(self, target, attacker, damage, dflags, mod):
        return super().dispatch(target, attacker, damage, dflags, mod)


EVENT_DISPATCHERS = EventDispatcherManager()
EVENT_DISPATCHERS.add_dispatcher(ConsolePrintDispatcher)
EVENT_DISPATCHERS.add_dispatcher(CommandDispatcher)
EVENT_DISPATCHERS.add_dispatcher(ClientCommandDispatcher)
EVENT_DISPATCHERS.add_dispatcher(ServerCommandDispatcher)
EVENT_DISPATCHERS.add_dispatcher(FrameEventDispatcher)
EVENT_DISPATCHERS.add_dispatcher(SetConfigstringDispatcher)
EVENT_DISPATCHERS.add_dispatcher(ChatEventDispatcher)
EVENT_DISPATCHERS.add_dispatcher(UnloadDispatcher)
EVENT_DISPATCHERS.add_dispatcher(PlayerConnectDispatcher)
EVENT_DISPATCHERS.add_dispatcher(PlayerLoadedDispatcher)
EVENT_DISPATCHERS.add_dispatcher(PlayerDisonnectDispatcher)
EVENT_DISPATCHERS.add_dispatcher(PlayerSpawnDispatcher)
EVENT_DISPATCHERS.add_dispatcher(KamikazeUseDispatcher)
EVENT_DISPATCHERS.add_dispatcher(KamikazeExplodeDispatcher)
EVENT_DISPATCHERS.add_dispatcher(PlayerItemsTossDispatcher)
EVENT_DISPATCHERS.add_dispatcher(StatsDispatcher)
EVENT_DISPATCHERS.add_dispatcher(VoteCalledDispatcher)
EVENT_DISPATCHERS.add_dispatcher(VoteStartedDispatcher)
EVENT_DISPATCHERS.add_dispatcher(VoteEndedDispatcher)
EVENT_DISPATCHERS.add_dispatcher(VoteDispatcher)
EVENT_DISPATCHERS.add_dispatcher(GameCountdownDispatcher)
EVENT_DISPATCHERS.add_dispatcher(GameStartDispatcher)
EVENT_DISPATCHERS.add_dispatcher(GameEndDispatcher)
EVENT_DISPATCHERS.add_dispatcher(RoundCountdownDispatcher)
EVENT_DISPATCHERS.add_dispatcher(RoundStartDispatcher)
EVENT_DISPATCHERS.add_dispatcher(RoundEndDispatcher)
EVENT_DISPATCHERS.add_dispatcher(TeamSwitchDispatcher)
EVENT_DISPATCHERS.add_dispatcher(TeamSwitchAttemptDispatcher)
EVENT_DISPATCHERS.add_dispatcher(MapDispatcher)
EVENT_DISPATCHERS.add_dispatcher(NewGameDispatcher)
EVENT_DISPATCHERS.add_dispatcher(KillDispatcher)
EVENT_DISPATCHERS.add_dispatcher(DeathDispatcher)
EVENT_DISPATCHERS.add_dispatcher(UserinfoDispatcher)
EVENT_DISPATCHERS.add_dispatcher(DamageDispatcher)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sys

    if sys.version_info >= (3, 11):
        from typing import NotRequired
    else:
        from typing_extensions import NotRequired
    from typing import Iterable, TypedDict
    from shinqlx import (
        PlayerInfo,
        PlayerState,
        PlayerStats,
        Vector3,
        Weapons,
        Powerups,
        Flight,
        AbstractChannel,
    )

_DUMMY_USERINFO: Iterable[str]

UserInfo = TypedDict(
    "UserInfo",
    {
        "ip": str,
        "ui_singlePlayerActive": str,
        "cg_autoAction": str,
        "cg_autoHop": str,
        "cg_predictItems": str,
        "model": str,
        "headmodel": str,
        "cl_anonymous": str,
        "country": str,
        "color1": str,
        "rate": str,
        "color2": str,
        "sex": str,
        "teamtask": str,
        "name": str,
        "handicap": str,
        "password": NotRequired[str],
    },
)

class NonexistentPlayerError(Exception): ...

class Player:
    @classmethod
    def all_players(cls) -> list[Player]: ...

    _valid: bool
    _id: int
    _info: PlayerInfo | None
    _userinfo: UserInfo | None
    _steam_id = int
    _name: str

    def __init__(self, client_id: int, info: PlayerInfo | None = ...) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __contains__(self, key: str) -> bool: ...
    def __getitem__(self, key: str) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def update(self) -> None: ...
    def _invalidate(self, e: str = ...) -> None: ...
    @property
    def cvars(self) -> dict[str, str | int]: ...
    @cvars.setter
    def cvars(self, new_cvars: dict[str, str | int]) -> None: ...
    @property
    def steam_id(self) -> int: ...
    @property
    def id(self) -> int: ...
    @property
    def ip(self) -> str: ...
    @property
    def clan(self) -> str: ...
    @clan.setter
    def clan(self, tag: str) -> None: ...
    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, value: str) -> None: ...
    @property
    def clean_name(self) -> str: ...
    @property
    def qport(self) -> int: ...
    @property
    def team(self) -> str: ...
    @team.setter
    def team(self, new_team: str) -> None: ...
    @property
    def colors(self) -> tuple[float, float]: ...
    @colors.setter
    def colors(self, value: tuple[float, float]) -> None: ...
    @property
    def model(self) -> str: ...
    @model.setter
    def model(self, value: str) -> None: ...
    @property
    def headmodel(self) -> str: ...
    @headmodel.setter
    def headmodel(self, value: str) -> None: ...
    @property
    def handicap(self) -> int: ...
    @handicap.setter
    def handicap(self, value: int) -> None: ...
    @property
    def autohop(self) -> bool: ...
    @autohop.setter
    def autohop(self, value: bool) -> None: ...
    @property
    def autoaction(self) -> bool: ...
    @autoaction.setter
    def autoaction(self, value: bool) -> None: ...
    @property
    def predictitems(self) -> bool: ...
    @predictitems.setter
    def predictitems(self, value: bool) -> None: ...
    @property
    def connection_state(self) -> int: ...
    @property
    def state(self) -> PlayerState | None: ...
    @property
    def privileges(self) -> str: ...
    @privileges.setter
    def privileges(self, value: None | str) -> None: ...
    @property
    def country(self) -> str: ...
    @country.setter
    def country(self, value: str) -> None: ...
    @property
    def valid(self) -> bool: ...
    @property
    def stats(self) -> PlayerStats | None: ...
    @property
    def ping(self) -> int: ...
    def position(self, reset: bool = ..., **kwargs: int) -> bool | Vector3: ...
    def velocity(self, reset: bool = ..., **kwargs: int) -> bool | Vector3: ...
    def weapons(self, reset: bool = ..., **kwargs: bool) -> bool | Weapons: ...
    def weapon(self, new_weapon: int | str | None = ...) -> bool | int: ...
    def ammo(self, reset: bool = ..., **kwargs: int) -> bool | Weapons: ...
    def powerups(self, reset: bool = ..., **kwargs: int) -> bool | Powerups: ...
    @property
    def holdable(self) -> int: ...
    @holdable.setter
    def holdable(self, value: str | None) -> None: ...
    def drop_holdable(self) -> None: ...
    def flight(self, reset: bool = ..., **kwargs: int) -> bool | Flight: ...
    @property
    def noclip(self) -> bool: ...
    @noclip.setter
    def noclip(self, value: bool | int | str) -> None: ...
    @property
    def health(self) -> int: ...
    @health.setter
    def health(self, value: int) -> None: ...
    @property
    def armor(self) -> int: ...
    @armor.setter
    def armor(self, value: int) -> None: ...
    @property
    def is_alive(self) -> bool: ...
    @is_alive.setter
    def is_alive(self, value: bool) -> None: ...
    @property
    def is_frozen(self) -> bool: ...
    @property
    def score(self) -> int: ...
    @score.setter
    def score(self, value: int) -> None: ...
    @property
    def channel(self) -> AbstractChannel: ...
    def center_print(self, msg: str) -> None: ...
    def tell(self, msg: str, **kwargs: str) -> None: ...
    def kick(self, reason: str = ...) -> None: ...
    def ban(self) -> None: ...
    def tempban(self) -> None: ...
    def addadmin(self) -> None: ...
    def addmod(self) -> None: ...
    def demote(self) -> None: ...
    def mute(self) -> None: ...
    def unmute(self) -> None: ...
    def put(self, team: str) -> None: ...
    def addscore(self, score: int) -> None: ...
    def switch(self, other_player: Player) -> None: ...
    def slap(self, damage: int = ...) -> None: ...
    def slay(self) -> None: ...
    def slay_with_mod(self, mod: int) -> bool: ...

class AbstractDummyPlayer(Player):
    def __init__(self, name: str = ...) -> None: ...
    @property
    def id(self) -> int: ...
    @property
    def steam_id(self) -> int: ...
    def update(self) -> None: ...
    @property
    def channel(self) -> AbstractChannel: ...
    def tell(self, msg: str, **kwargs: str) -> None: ...

class RconDummyPlayer(AbstractDummyPlayer):
    def __init__(self) -> None: ...
    @property
    def steam_id(self) -> int: ...
    @property
    def channel(self) -> AbstractChannel: ...
    def tell(self, msg: str, **kwargs: str) -> None: ...



from datetime import timedelta
import redis

import shinqlx


# ====================================================================
#                          AbstractDatabase
# ====================================================================
# noinspection PyProtectedMember
class AbstractDatabase:
    # An instance counter. Useful for closing connections.
    _counter = 0

    def __init__(self, plugin):
        self.plugin = plugin
        self.__class__._counter += 1

    def __del__(self):
        self.__class__._counter -= 1

    @property
    def logger(self):
        return shinqlx.get_logger(self.plugin)

    def set_permission(self, player, level):
        """Abstract method. Should set the permission of a player.

        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def get_permission(self, player):
        """Abstract method. Should return the permission of a player.

        :returns: int
        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def has_permission(self, player, level=5):
        """Abstract method. Should return whether or not a player has more than or equal
        to a certain permission level. Should only take a value of 0 to 5, where 0 is
        always True.

        :returns: bool
        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def set_flag(self, player, flag, value=True):
        """Abstract method. Should set specified player flag to value.

        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def clear_flag(self, player, flag):
        """Should clear specified player flag."""
        return self.set_flag(player, flag, False)

    def get_flag(self, player, flag, default=False):
        """Abstract method. Should return specified player flag

        :returns: bool
        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def connect(self):
        """Abstract method. Should return a connection to the database. Exactly what a
        "connection" obviously depends on the database, so the specifics will be up
        to the implementation.

        A :class:`shinqlx.Plugin` subclass can set

        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")

    def close(self):
        """Abstract method. If the database has a connection state, this method should
        close the connection.

        :raises: NotImplementedError

        """
        raise NotImplementedError("The base plugin can't do database actions.")


# ====================================================================
#                               Redis
# ====================================================================
# noinspection PyProtectedMember
class Redis(AbstractDatabase):
    """A subclass of :class:`shinqlx.AbstractDatabase` providing support for Redis."""

    # We only use the instance-level ones if we override the URI from the config.
    _conn = None
    _pool = None
    _pass = ""

    def __del__(self):
        super().__del__()
        self.close()

    def __contains__(self, key):
        return self.r.exists(key)

    def __getitem__(self, key):
        res = self.r.get(key)
        if res is None:
            raise KeyError(f"The key '{key}' is not present in the database.")
        return res

    def __setitem__(self, key, item):
        res = self.r.set(key, item)
        if res is False:
            raise RuntimeError("The database assignment failed.")

    def __delitem__(self, key):
        res = self.r.delete(key)
        if res == 0:
            raise KeyError(f"The key '{key}' is not present in the database.")

    def __getattr__(self, attr):
        return getattr(self.r, attr)

    @property
    def r(self):
        return self.connect()

    def set_permission(self, player, level):
        """Sets the permission of a player.

        :param: player: The player in question.
        :type: player: shinqlx.Player

        """
        if isinstance(player, shinqlx.Player):
            key = f"minqlx:players:{player.steam_id}:permission"
        else:
            key = f"minqlx:players:{player}:permission"

        self[key] = level

    def get_permission(self, player):
        """Gets the permission of a player.

        :param: player: The player in question.
        :type: player: shinqlx.Player, int
        :returns: int

        """
        if isinstance(player, shinqlx.Player):
            steam_id = player.steam_id
        elif isinstance(player, int):
            steam_id = player
        elif isinstance(player, str):
            steam_id = int(player)
        else:
            raise ValueError(
                "Invalid player. Use either a shinqlx.Player instance or a SteamID64."
            )

        # If it's the owner, treat it like a 5.
        if steam_id == shinqlx.owner():
            return 5

        key = f"minqlx:players:{steam_id}:permission"
        try:
            perm = self[key]
        except KeyError:
            perm = "0"

        return int(perm)

    def has_permission(self, player, level=5):
        """Checks if the player has higher than or equal to *level*.

        :param: player: The player in question.
        :type: player: shinqlx.Player
        :param: level: The permission level to check for.
        :type: level: int
        :returns: bool

        """
        return self.get_permission(player) >= level

    def set_flag(self, player, flag, value=True):
        """Sets specified player flag

        :param: player: The player in question.
        :type: player: shinqlx.Player
        :param: flag: The flag to set.
        :type: flag: string
        :param: value: (optional, default=True) Value to set
        :type: value: bool

        """
        if isinstance(player, shinqlx.Player):
            key = f"minqlx:players:{player.steam_id}:flags:{flag}"
        else:
            key = f"minqlx:players:{player}:flags:{flag}"

        self[key] = 1 if value else 0

    def get_flag(self, player, flag, default=False):
        """Clears the specified player flag

        :param: player: The player in question.
        :type: player: shinqlx.Player
        :param: flag: The flag to get
        :type: flag: string
        :param: default: (optional, default=False) The value to return if the flag is unknown
        :type: default: bool

        """
        if isinstance(player, shinqlx.Player):
            key = f"minqlx:players:{player.steam_id}:flags:{flag}"
        else:
            key = f"minqlx:players:{player}:flags:{flag}"

        try:
            return bool(int(self[key]))
        except KeyError:
            return default

    def connect(self, host=None, database=0, unix_socket=False, password=None):
        """Returns a connection to a Redis database. If *host* is None, it will
        fall back to the settings in the config and ignore the rest of the arguments.
        It will also share the connection across any plugins using the default
        configuration. Passing *host* will make it connect to a specific database
        that is not shared at all. Subsequent calls to this will return the connection
        initialized the first call unless it has been closed.

        :param: host: The host name. If no port is specified, it will use 6379. Ex.: ``localhost:1234``.
        :type: host: str
        :param: database: The database number that should be used.
        :type: database: int
        :param: unix_socket: Whether or not *host* should be interpreted as a unix socket path.
        :type: unix_socket: bool
        :raises: RuntimeError

        """
        if not host and not self._conn:  # Resort to default settings in config?
            if not Redis._conn:
                cvar_host = shinqlx.get_cvar("qlx_redisAddress")
                if cvar_host is None:
                    raise ValueError("cvar qlx_redisAddress misconfigured")
                redis_db_cvar = shinqlx.get_cvar("qlx_redisDatabase")
                if redis_db_cvar is None:
                    raise ValueError("cvar qlx_redisDatabase misconfigured.")
                cvar_db = int(redis_db_cvar)
                unix_socket_cvar = shinqlx.get_cvar("qlx_redisUnixSocket")
                if unix_socket_cvar is None:
                    raise ValueError("cvar qlx_redisUnixSocket misconfigured")
                cvar_unixsocket = bool(int(unix_socket_cvar))
                password_cvar = shinqlx.get_cvar("qlx_redisPassword")
                if password_cvar is None:
                    raise ValueError("cvar qlx_redisPassword misconfigured")
                Redis._pass = password_cvar
                if cvar_unixsocket:
                    Redis._conn = redis.StrictRedis(
                        unix_socket_path=cvar_host,
                        db=cvar_db,
                        password=Redis._pass,
                        decode_responses=True,
                    )
                else:
                    split_host = cvar_host.split(":")
                    port = int(split_host[1]) if len(split_host) > 1 else 6379
                    Redis._pool = redis.ConnectionPool(
                        host=split_host[0],
                        port=port,
                        db=cvar_db,
                        password=Redis._pass,
                        decode_responses=True,
                    )
                    Redis._conn = redis.StrictRedis(
                        connection_pool=Redis._pool, decode_responses=True
                    )
                    # TODO: Why does self._conn get set when doing Redis._conn?
                    self._conn = None
            return Redis._conn
        if not self._conn:
            if host is None:
                raise ValueError("wrong host")
            split_host = host.split(":")
            port = int(split_host[1]) if len(split_host) > 1 else 6379

            if unix_socket:
                self._conn = redis.StrictRedis(
                    unix_socket_path=host,
                    db=database,
                    password=password,
                    decode_responses=True,
                )
            else:
                self._pool = redis.ConnectionPool(
                    host=split_host[0],
                    port=port,
                    db=database,
                    password=password,
                    decode_responses=True,
                )
                self._conn = redis.StrictRedis(
                    connection_pool=self._pool, decode_responses=True
                )
        return self._conn

    def close(self):
        """Close the Redis connection if the config was overridden. Otherwise only do so
        if this is the last plugin using the default connection.

        """
        if self._conn:
            self._conn = None
            if self._pool:
                self._pool.disconnect()
                self._pool = None

        if Redis._counter <= 1 and Redis._conn:
            Redis._conn = None
            if Redis._pool:
                Redis._pool.disconnect()
                Redis._pool = None

    def mset(self, *args, **kwargs):
        mapping = {}
        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise redis.RedisError("MSET requires **kwargs or a single dict arg")
            mapping.update(args[0])

        if kwargs:
            mapping.update(kwargs)

        return self.r.mset(mapping)

    def msetnx(self, *args, **kwargs):
        mapping = {}
        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise redis.RedisError("MSETNX requires **kwargs or a single dict arg")
            mapping.update(args[0])

        if kwargs:
            mapping.update(kwargs)

        return self.r.msetnx(mapping)

    def zadd(self, name, *args, **kwargs):
        if redis.VERSION < (3, 0):
            return self.r.zadd(name, *args, **kwargs)

        if isinstance(args[0], dict):
            return self.r.zadd(name, *args, **kwargs)

        mapping = {}
        if len(args) > 0 and len(args) % 2 != 0:
            raise redis.RedisError("ZADD requires an equal number of values and scores")

        for i in range(0, len(args), 2):
            mapping[args[i + 1]] = args[i]

        return self.r.zadd(name, mapping, **kwargs)

    def zincrby(self, name, value_or_amount, amount_or_value=1):
        if not isinstance(value_or_amount, (int, float)):
            value = value_or_amount
            amount = amount_or_value
        else:
            value = amount_or_value
            amount = value_or_amount

        if redis.VERSION < (3, 0):
            return self.r.zincrby(name, value, amount)
        return self.r.zincrby(name, amount, value)

    def setex(self, name, value_or_time, time_or_value):
        if not isinstance(value_or_time, (int, timedelta)):
            value = value_or_time
            time = time_or_value
        else:
            value = time_or_value
            time = value_or_time

        if redis.VERSION < (3, 0):
            return self.r.setex(name, time, value)
        return self.r.setex(name, value, time)

    def lrem(self, name, value_or_count, num_or_value=0):
        if not isinstance(value_or_count, int):
            value = value_or_count
            count = num_or_value
        else:
            value = num_or_value
            count = value_or_count

        if redis.VERSION < (3, 0):
            return self.r.lrem(name, count, value)
        return self.r.lrem(name, value, count)



from abc import abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Pattern, Callable, Iterable
    from shinqlx import Player, Plugin

MAX_MSG_LENGTH: int
re_color_tag: Pattern

class AbstractChannel:
    _name: str

    def __init__(self, name: str) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    @property
    def name(self) -> str: ...
    def reply(self, msg: str, limit: int = ..., delimiter: str = ...) -> None: ...
    def split_long_lines(
        self, msg: str, limit: int = ..., delimiter: str = ...
    ) -> list[str]: ...

class ChatChannel(AbstractChannel):
    fmt: str

    def __init__(self, name: str = ..., fmt: str = ...) -> None: ...
    @abstractmethod
    def receipients(self) -> list[int] | None: ...
    def reply(self, msg: str, limit: int = ..., delimiter: str = ...) -> None: ...

class TeamChatChannel(ChatChannel):
    team: str
    def __init__(self, team: str = ..., name: str = ..., fmt: str = ...) -> None: ...
    def receipients(self) -> list[int] | None: ...

class TellChannel(ChatChannel):
    recipient: str | int | Player

    def __init__(self, player: str | int | Player) -> None: ...
    def __repr__(self) -> str: ...
    def receipients(self) -> list[int] | None: ...

class ConsoleChannel(AbstractChannel):
    def __init__(self) -> None: ...
    def reply(self, msg: str, limit: int = ..., delimiter: str = ...) -> None: ...

class ClientCommandChannel(AbstractChannel):
    recipient: Player
    tell_channel: ChatChannel

    def __init__(self, player: Player) -> None: ...
    def __repr__(self) -> str: ...
    def reply(self, msg: str, limit: int = ..., delimiter: str = ...) -> None: ...

class Command:
    name: list[str]
    plugin: Plugin
    handler: Callable
    permission: int
    channels: list[AbstractChannel]
    exclude_channels: list[AbstractChannel]
    client_cmd_pass: bool
    client_cmd_perm: int
    prefix: bool
    usage: str

    def __init__(
        self,
        plugin: Plugin,
        name: str | Iterable[str],
        handler: Callable,
        permission: int,
        channels: Iterable[AbstractChannel] | None,
        exclude_channels: Iterable[AbstractChannel] | None,
        client_cmd_pass: bool,
        client_cmd_perm: int,
        prefix: bool,
        usage: str,
    ) -> None: ...
    def execute(
        self, player: Player, msg: str, channel: AbstractChannel
    ) -> int | None: ...
    def is_eligible_name(self, name: str) -> bool: ...
    def is_eligible_channel(self, channel: AbstractChannel) -> bool: ...
    def is_eligible_player(self, player: Player, is_client_cmd: bool) -> bool: ...

class CommandInvoker:
    _commands: tuple[
        list[Command], list[Command], list[Command], list[Command], list[Command]
    ]

    def __init__(self) -> None: ...
    @property
    def commands(self) -> list[Command]: ...
    def add_command(self, command: Command, priority: int) -> None: ...
    def remove_command(self, command: Command) -> None: ...
    def is_registered(self, command: Command) -> bool: ...
    def handle_input(
        self, player: Player, msg: str, channel: AbstractChannel
    ) -> bool: ...

COMMANDS: CommandInvoker
CHAT_CHANNEL: AbstractChannel
RED_TEAM_CHAT_CHANNEL: AbstractChannel
BLUE_TEAM_CHAT_CHANNEL: AbstractChannel
FREE_CHAT_CHANNEL: AbstractChannel
SPECTATOR_CHAT_CHANNEL: AbstractChannel
CONSOLE_CHANNEL: AbstractChannel



from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from typing import Type, Callable, Iterable, Pattern, Mapping
    from shinqlx import (
        Plugin,
        Player,
        AbstractChannel,
        Command,
        StatsData,
        GameStartData,
        GameEndData,
        RoundEndData,
        KillData,
        DeathData,
        UserInfoEventInput,
    )

_re_vote: Pattern
hot_plugged_events: Iterable[str]

UncancellableEventReturn = Literal["RET_NONE"] | None
CancellableEventReturn = Literal[0, 1, 2, 3] | None

class EventDispatcher:
    name: str
    plugins: dict[
        Plugin,
        tuple[
            Iterable[Callable],
            Iterable[Callable],
            Iterable[Callable],
            Iterable[Callable],
            Iterable[Callable],
        ],
    ]
    _args: Iterable[str] | None
    _kwargs: Mapping[str, str] | None
    _return_value: str | bool | Iterable | None
    no_debug: Iterable[str]
    need_zmq_stats_enabled: bool

    def __init__(self) -> None: ...
    @property
    def args(self) -> Iterable[str]: ...
    @args.setter
    def args(self, value: Iterable[str]) -> None: ...
    @property
    def kwargs(self) -> Mapping[str, str]: ...
    @kwargs.setter
    def kwargs(self, value: Mapping[str, str]) -> None: ...
    @property
    def return_value(self) -> str | bool | Iterable | None: ...
    @return_value.setter
    def return_value(self, value: str | bool | Iterable | None) -> None: ...
    def dispatch(self, *args, **kwargs) -> str | bool | Iterable | None: ...  # type: ignore
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...
    def add_hook(
        self, plugin: Plugin | str, handler: Callable, priority: int = ...
    ) -> None: ...
    def remove_hook(
        self, plugin: Plugin | str, handler: Callable, priority: int = ...
    ) -> None: ...

class EventDispatcherManager:
    _dispatchers: dict[str, EventDispatcher]

    def __init__(self) -> None: ...
    def __getitem__(self, key: str) -> EventDispatcher: ...
    def __contains__(self, key: str) -> bool: ...
    def add_dispatcher(self, dispatcher: Type[EventDispatcher]) -> None: ...
    def remove_dispatcher(self, dispatcher: Type[EventDispatcher]) -> None: ...
    def remove_dispatcher_by_name(self, event_name: str) -> None: ...

class ConsolePrintDispatcher(EventDispatcher):
    def dispatch(self, text: str) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class CommandDispatcher(EventDispatcher):
    def dispatch(self, caller: Player, command: Command, args: str) -> None: ...

class ClientCommandDispatcher(EventDispatcher):
    def dispatch(self, player: Player, cmd: str) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class ServerCommandDispatcher(EventDispatcher):
    def dispatch(self, player: Player, cmd: str) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class FrameEventDispatcher(EventDispatcher):
    def dispatch(self) -> str | bool | Iterable | None: ...

class SetConfigstringDispatcher(EventDispatcher):
    def dispatch(self, index: int, value: str) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class ChatEventDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, msg: str, channel: AbstractChannel
    ) -> str | bool | Iterable | None: ...

class UnloadDispatcher(EventDispatcher):
    def dispatch(self, plugin: Plugin) -> None: ...

class PlayerConnectDispatcher(EventDispatcher):
    def dispatch(self, player: Player) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class PlayerLoadedDispatcher(EventDispatcher):
    def dispatch(self, player: Player) -> str | bool | Iterable | None: ...

class PlayerDisonnectDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, reason: str | None
    ) -> str | bool | Iterable | None: ...

class PlayerSpawnDispatcher(EventDispatcher):
    def dispatch(self, player: Player) -> str | bool | Iterable | None: ...

class StatsDispatcher(EventDispatcher):
    def dispatch(self, stats: StatsData) -> str | bool | Iterable | None: ...

class VoteCalledDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, vote: str, args: str | None
    ) -> str | bool | Iterable | None: ...

class VoteStartedDispatcher(EventDispatcher):
    _caller: Player | None

    def __init__(self) -> None: ...
    def dispatch(self, vote: str, args: str | None) -> str | bool | Iterable | None: ...
    def caller(self, player: Player) -> None: ...

class VoteEndedDispatcher(EventDispatcher):
    def dispatch(self, passed: bool) -> None: ...

class VoteDispatcher(EventDispatcher):
    def dispatch(self, player: Player, yes: bool) -> str | bool | Iterable | None: ...

class GameCountdownDispatcher(EventDispatcher):
    def dispatch(self) -> str | bool | Iterable | None: ...

class GameStartDispatcher(EventDispatcher):
    def dispatch(self, data: GameStartData) -> str | bool | Iterable | None: ...

class GameEndDispatcher(EventDispatcher):
    def dispatch(self, data: GameEndData) -> str | bool | Iterable | None: ...

class RoundCountdownDispatcher(EventDispatcher):
    def dispatch(self, round_number: int) -> str | bool | Iterable | None: ...

class RoundStartDispatcher(EventDispatcher):
    def dispatch(self, round_number: int) -> str | bool | Iterable | None: ...

class RoundEndDispatcher(EventDispatcher):
    def dispatch(self, data: RoundEndData) -> str | bool | Iterable | None: ...

class TeamSwitchDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, old_team: str, new_team: str
    ) -> str | bool | Iterable | None: ...

class TeamSwitchAttemptDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, old_team: str, new_team: str
    ) -> str | bool | Iterable | None: ...

class MapDispatcher(EventDispatcher):
    def dispatch(self, mapname: str, factory: str) -> str | bool | Iterable | None: ...

class NewGameDispatcher(EventDispatcher):
    def dispatch(self) -> str | bool | Iterable | None: ...

class KillDispatcher(EventDispatcher):
    def dispatch(
        self, victim: Player, killer: Player | None, data: KillData
    ) -> str | bool | Iterable | None: ...

class DeathDispatcher(EventDispatcher):
    def dispatch(
        self, victim: Player, killer: Player | None, data: DeathData
    ) -> str | bool | Iterable | None: ...

class UserinfoDispatcher(EventDispatcher):
    def dispatch(
        self, playe: Player, changed: UserInfoEventInput
    ) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class KamikazeUseDispatcher(EventDispatcher):
    def dispatch(self, player: Player) -> str | bool | Iterable | None: ...

class KamikazeExplodeDispatcher(EventDispatcher):
    def dispatch(
        self, player: Player, is_used_on_demand: bool
    ) -> str | bool | Iterable | None: ...

class PlayerItemsTossDispatcher(EventDispatcher):
    def dispatch(self, player: Player) -> str | bool | Iterable | None: ...
    def handle_return(
        self, handler: Callable, value: int | str | None
    ) -> str | None: ...

class DamageDispatcher(EventDispatcher):
    def dispatch(
        self,
        target: Player | int | None,
        attacker: Player | int | None,
        damage: int,
        dflags: int,
        means_of_death: int,
    ) -> str | bool | Iterable | None: ...

EVENT_DISPATCHERS: EventDispatcherManager



from ._shinqlx import (
    DEBUG,
    RET_NONE,
    RET_STOP,
    RET_STOP_EVENT,
    RET_STOP_ALL,
    RET_USAGE,
    PRI_HIGHEST,
    PRI_HIGH,
    PRI_NORMAL,
    PRI_LOW,
    PRI_LOWEST,
    CVAR_ARCHIVE,
    CVAR_USERINFO,
    CVAR_SERVERINFO,
    CVAR_SYSTEMINFO,
    CVAR_INIT,
    CVAR_LATCH,
    CVAR_ROM,
    CVAR_USER_CREATED,
    CVAR_TEMP,
    CVAR_CHEAT,
    CVAR_NORESTART,
    PRIV_NONE,
    PRIV_MOD,
    PRIV_ADMIN,
    PRIV_ROOT,
    PRIV_BANNED,
    CS_FREE,
    CS_ZOMBIE,
    CS_CONNECTED,
    CS_PRIMED,
    CS_ACTIVE,
    TEAM_FREE,
    TEAM_RED,
    TEAM_BLUE,
    TEAM_SPECTATOR,
    MOD_UNKNOWN,
    MOD_SHOTGUN,
    MOD_GAUNTLET,
    MOD_MACHINEGUN,
    MOD_GRENADE,
    MOD_GRENADE_SPLASH,
    MOD_ROCKET,
    MOD_ROCKET_SPLASH,
    MOD_PLASMA,
    MOD_PLASMA_SPLASH,
    MOD_RAILGUN,
    MOD_LIGHTNING,
    MOD_BFG,
    MOD_BFG_SPLASH,
    MOD_WATER,
    MOD_SLIME,
    MOD_LAVA,
    MOD_CRUSH,
    MOD_TELEFRAG,
    MOD_FALLING,
    MOD_SUICIDE,
    MOD_TARGET_LASER,
    MOD_TRIGGER_HURT,
    MOD_NAIL,
    MOD_CHAINGUN,
    MOD_PROXIMITY_MINE,
    MOD_KAMIKAZE,
    MOD_JUICED,
    MOD_GRAPPLE,
    MOD_SWITCH_TEAMS,
    MOD_THAW,
    MOD_LIGHTNING_DISCHARGE,
    MOD_HMG,
    MOD_RAILGUN_HEADSHOT,
    DAMAGE_RADIUS,
    DAMAGE_NO_ARMOR,
    DAMAGE_NO_KNOCKBACK,
    DAMAGE_NO_PROTECTION,
    DAMAGE_NO_TEAM_PROTECTION,
    Vector3,
    Flight,
    Powerups,
    Weapons,
    PlayerInfo,
    PlayerState,
    PlayerStats,
    player_info,
    players_info,
    get_userinfo,
    send_server_command,
    client_command,
    console_command,
    get_cvar,
    set_cvar,
    set_cvar_limit,
    kick,
    console_print,
    get_configstring,
    set_configstring,
    force_vote,
    add_console_command,
    player_state,
    player_stats,
    set_position,
    set_velocity,
    noclip,
    set_health,
    set_armor,
    set_weapons,
    set_weapon,
    set_ammo,
    set_powerups,
    set_holdable,
    drop_holdable,
    set_flight,
    set_invulnerability,
    set_score,
    callvote,
    allow_single_player,
    player_spawn,
    set_privileges,
    destroy_kamikaze_timers,
    spawn_item,
    remove_dropped_items,
    slay_with_mod,
    replace_items,
    dev_print_items,
    force_weapon_respawn_time,
    register_handler,
    get_targetting_entities,
)
from ._core import (
    PluginLoadError,
    PluginUnloadError,
    TEAMS,
    GAMETYPES,
    GAMETYPES_SHORT,
    CONNECTION_STATES,
    WEAPONS,
    DEFAULT_PLUGINS,
    parse_variables,
    get_logger,
    log_exception,
    handle_exception,
    threading_excepthook,
    uptime,
    owner,
    stats_listener,
    set_cvar_once,
    set_cvar_limit_once,
    set_plugins_version,
    set_map_subtitles,
    next_frame,
    delay,
    thread,
    load_preset_plugins,
    load_plugin,
    unload_plugin,
    reload_plugin,
    initialize_cvars,
    initialize,
    late_init,
)
from ._game import Game, NonexistentGameError
from ._player import (
    Player,
    NonexistentPlayerError,
    AbstractDummyPlayer,
    RconDummyPlayer,
    UserInfo,
)
from ._plugin import (
    Plugin,
    GameStartData,
    GameEndData,
    RoundEndData,
    DeathData,
    KillData,
    UserInfoEventInput,
    Weapon,
    PowerUps,
    MeansOfDeath,
    PlayerSummaryData,
    Vector,
    Holdable,
    PlayerData,
    PlayerKillStats,
    PlayerDeathStats,
    PlayerMedalStats,
    MedalData,
    RoundOverStats,
    PlayerConnectStats,
    PlayerDisconnectStats,
    PlayerGameData,
    PlayerSwitchTeamStats,
    TeamSwitchGameData,
    MatchStartedStats,
    MatchReportStats,
    PlayerStatsStats,
    PlayerStatsEntry,
    WeaponsStatsEntry,
    PickupsEntry,
    MedalsEntry,
    DamageEntry,
    SingleWeaponStatsEntry,
    StatsData,
)
from ._events import (
    EventDispatcher,
    EventDispatcherManager,
    ConsolePrintDispatcher,
    CommandDispatcher,
    ClientCommandDispatcher,
    ServerCommandDispatcher,
    FrameEventDispatcher,
    SetConfigstringDispatcher,
    ChatEventDispatcher,
    UnloadDispatcher,
    PlayerConnectDispatcher,
    PlayerLoadedDispatcher,
    PlayerDisonnectDispatcher,
    PlayerSpawnDispatcher,
    StatsDispatcher,
    VoteCalledDispatcher,
    VoteStartedDispatcher,
    VoteEndedDispatcher,
    VoteDispatcher,
    GameCountdownDispatcher,
    GameStartDispatcher,
    GameEndDispatcher,
    RoundCountdownDispatcher,
    RoundStartDispatcher,
    RoundEndDispatcher,
    TeamSwitchDispatcher,
    TeamSwitchAttemptDispatcher,
    MapDispatcher,
    NewGameDispatcher,
    KillDispatcher,
    DeathDispatcher,
    UserinfoDispatcher,
    KamikazeUseDispatcher,
    KamikazeExplodeDispatcher,
    PlayerItemsTossDispatcher,
    DamageDispatcher,
    EVENT_DISPATCHERS,
    UncancellableEventReturn,
    CancellableEventReturn,
)
from ._handlers import (
    frame_tasks,
    next_frame_tasks,
    handle_rcon,
    handle_client_command,
    handle_server_command,
    handle_frame,
    handle_new_game,
    handle_set_configstring,
    handle_player_connect,
    handle_player_loaded,
    handle_player_disconnect,
    handle_player_spawn,
    handle_kamikaze_use,
    handle_kamikaze_explode,
    handle_console_print,
    handle_damage,
    redirect_print,
    register_handlers,
)
from ._commands import (
    MAX_MSG_LENGTH,
    re_color_tag,
    AbstractChannel,
    ChatChannel,
    TeamChatChannel,
    TellChannel,
    ConsoleChannel,
    ClientCommandChannel,
    Command,
    CommandInvoker,
    COMMANDS,
    CHAT_CHANNEL,
    RED_TEAM_CHAT_CHANNEL,
    BLUE_TEAM_CHAT_CHANNEL,
    FREE_CHAT_CHANNEL,
    SPECTATOR_CHAT_CHANNEL,
    CONSOLE_CHANNEL,
)
from ._zmq import StatsListener

__version__: str
__plugins_version__: str

_map_title: str | None
_map_subtitle1: str | None
_map_subtitle2: str | None

__version_info__: tuple[int, int, int]

__all__ = [
    "__version__",
    "__version_info__",
    "__plugins_version__",
    # _shinqlx
    "DEBUG",
    "RET_NONE",
    "RET_STOP",
    "RET_STOP_EVENT",
    "RET_STOP_ALL",
    "RET_USAGE",
    "PRI_HIGHEST",
    "PRI_HIGH",
    "PRI_NORMAL",
    "PRI_LOW",
    "PRI_LOWEST",
    "CVAR_ARCHIVE",
    "CVAR_USERINFO",
    "CVAR_SERVERINFO",
    "CVAR_SYSTEMINFO",
    "CVAR_INIT",
    "CVAR_LATCH",
    "CVAR_ROM",
    "CVAR_USER_CREATED",
    "CVAR_TEMP",
    "CVAR_CHEAT",
    "CVAR_NORESTART",
    "PRIV_NONE",
    "PRIV_MOD",
    "PRIV_ADMIN",
    "PRIV_ROOT",
    "PRIV_BANNED",
    "CS_FREE",
    "CS_ZOMBIE",
    "CS_CONNECTED",
    "CS_PRIMED",
    "CS_ACTIVE",
    "TEAM_FREE",
    "TEAM_RED",
    "TEAM_BLUE",
    "TEAM_SPECTATOR",
    "MOD_UNKNOWN",
    "MOD_SHOTGUN",
    "MOD_GAUNTLET",
    "MOD_MACHINEGUN",
    "MOD_GRENADE",
    "MOD_GRENADE_SPLASH",
    "MOD_ROCKET",
    "MOD_ROCKET_SPLASH",
    "MOD_PLASMA",
    "MOD_PLASMA_SPLASH",
    "MOD_RAILGUN",
    "MOD_LIGHTNING",
    "MOD_BFG",
    "MOD_BFG_SPLASH",
    "MOD_WATER",
    "MOD_SLIME",
    "MOD_LAVA",
    "MOD_CRUSH",
    "MOD_TELEFRAG",
    "MOD_FALLING",
    "MOD_SUICIDE",
    "MOD_TARGET_LASER",
    "MOD_TRIGGER_HURT",
    "MOD_NAIL",
    "MOD_CHAINGUN",
    "MOD_PROXIMITY_MINE",
    "MOD_KAMIKAZE",
    "MOD_JUICED",
    "MOD_GRAPPLE",
    "MOD_SWITCH_TEAMS",
    "MOD_THAW",
    "MOD_LIGHTNING_DISCHARGE",
    "MOD_HMG",
    "MOD_RAILGUN_HEADSHOT",
    "DAMAGE_RADIUS",
    "DAMAGE_NO_ARMOR",
    "DAMAGE_NO_KNOCKBACK",
    "DAMAGE_NO_PROTECTION",
    "DAMAGE_NO_TEAM_PROTECTION",
    "Vector3",
    "Flight",
    "Powerups",
    "Weapons",
    "PlayerInfo",
    "PlayerState",
    "PlayerStats",
    "player_info",
    "players_info",
    "get_userinfo",
    "send_server_command",
    "client_command",
    "console_command",
    "get_cvar",
    "set_cvar",
    "set_cvar_limit",
    "kick",
    "console_print",
    "get_configstring",
    "set_configstring",
    "force_vote",
    "add_console_command",
    "player_state",
    "player_stats",
    "set_position",
    "set_velocity",
    "noclip",
    "set_health",
    "set_armor",
    "set_weapons",
    "set_weapon",
    "set_ammo",
    "set_powerups",
    "set_holdable",
    "drop_holdable",
    "set_flight",
    "set_invulnerability",
    "set_score",
    "callvote",
    "allow_single_player",
    "player_spawn",
    "set_privileges",
    "destroy_kamikaze_timers",
    "spawn_item",
    "remove_dropped_items",
    "slay_with_mod",
    "replace_items",
    "dev_print_items",
    "force_weapon_respawn_time",
    "register_handler",
    "get_targetting_entities",
    # _core
    "PluginLoadError",
    "PluginUnloadError",
    "TEAMS",
    "GAMETYPES",
    "GAMETYPES_SHORT",
    "CONNECTION_STATES",
    "WEAPONS",
    "DEFAULT_PLUGINS",
    "parse_variables",
    "get_logger",
    "log_exception",
    "handle_exception",
    "threading_excepthook",
    "uptime",
    "owner",
    "stats_listener",
    "set_cvar_once",
    "set_cvar_limit_once",
    "set_plugins_version",
    "set_map_subtitles",
    "next_frame",
    "delay",
    "thread",
    "load_preset_plugins",
    "load_plugin",
    "unload_plugin",
    "reload_plugin",
    "initialize_cvars",
    "initialize",
    "late_init",
    # _plugin
    "Plugin",
    "GameStartData",
    "GameEndData",
    "RoundEndData",
    "DeathData",
    "KillData",
    "UserInfoEventInput",
    "Weapon",
    "PowerUps",
    "MeansOfDeath",
    "PlayerSummaryData",
    "Vector",
    "Holdable",
    "PlayerData",
    "PlayerKillStats",
    "PlayerDeathStats",
    "PlayerMedalStats",
    "MedalData",
    "RoundOverStats",
    "PlayerConnectStats",
    "PlayerDisconnectStats",
    "PlayerGameData",
    "PlayerSwitchTeamStats",
    "TeamSwitchGameData",
    "MatchStartedStats",
    "MatchReportStats",
    "PlayerStatsStats",
    "PlayerStatsEntry",
    "WeaponsStatsEntry",
    "PickupsEntry",
    "MedalsEntry",
    "DamageEntry",
    "SingleWeaponStatsEntry",
    "StatsData",
    # _game
    "Game",
    "NonexistentGameError",
    # _player
    "Player",
    "NonexistentPlayerError",
    "AbstractDummyPlayer",
    "RconDummyPlayer",
    "UserInfo",
    # _commands
    "MAX_MSG_LENGTH",
    "re_color_tag",
    "AbstractChannel",
    "ChatChannel",
    "TeamChatChannel",
    "TellChannel",
    "ConsoleChannel",
    "ClientCommandChannel",
    "Command",
    "CommandInvoker",
    "COMMANDS",
    "CHAT_CHANNEL",
    "RED_TEAM_CHAT_CHANNEL",
    "BLUE_TEAM_CHAT_CHANNEL",
    "FREE_CHAT_CHANNEL",
    "SPECTATOR_CHAT_CHANNEL",
    "CONSOLE_CHANNEL",
    # _events
    "EventDispatcher",
    "EventDispatcherManager",
    "ConsolePrintDispatcher",
    "CommandDispatcher",
    "ClientCommandDispatcher",
    "ServerCommandDispatcher",
    "FrameEventDispatcher",
    "SetConfigstringDispatcher",
    "ChatEventDispatcher",
    "UnloadDispatcher",
    "PlayerConnectDispatcher",
    "PlayerLoadedDispatcher",
    "PlayerDisonnectDispatcher",
    "PlayerSpawnDispatcher",
    "StatsDispatcher",
    "VoteCalledDispatcher",
    "VoteStartedDispatcher",
    "VoteEndedDispatcher",
    "VoteDispatcher",
    "GameCountdownDispatcher",
    "GameStartDispatcher",
    "GameEndDispatcher",
    "RoundCountdownDispatcher",
    "RoundStartDispatcher",
    "RoundEndDispatcher",
    "TeamSwitchDispatcher",
    "TeamSwitchAttemptDispatcher",
    "MapDispatcher",
    "NewGameDispatcher",
    "KillDispatcher",
    "DeathDispatcher",
    "UserinfoDispatcher",
    "KamikazeUseDispatcher",
    "KamikazeExplodeDispatcher",
    "PlayerItemsTossDispatcher",
    "DamageDispatcher",
    "EVENT_DISPATCHERS",
    "UncancellableEventReturn",
    "CancellableEventReturn",
    # _handlers
    "frame_tasks",
    "next_frame_tasks",
    "handle_rcon",
    "handle_client_command",
    "handle_server_command",
    "handle_frame",
    "handle_new_game",
    "handle_set_configstring",
    "handle_player_connect",
    "handle_player_loaded",
    "handle_player_disconnect",
    "handle_player_spawn",
    "handle_kamikaze_use",
    "handle_kamikaze_explode",
    "handle_console_print",
    "handle_damage",
    "redirect_print",
    "register_handlers",
    # _zmq
    "StatsListener",
]



from typing import TYPE_CHECKING
from threading import Thread

if TYPE_CHECKING:
    from typing import TypedDict
    from shinqlx import StatsData, GameStartData, GameEndData, RoundEndData, DeathData

TeamSwitchPlayerData = TypedDict(
    "TeamSwitchPlayerData", {"NAME": str, "OLD_TEAM": str, "STEAM_ID": str, "TEAM": str}
)

TeamSwitchData = TypedDict(
    "TeamSwitchData",
    {
        "KILLER": TeamSwitchPlayerData,
        "MATCH_GUID": str,
        "TIME": int,
        "WARMUP": bool,
    },
)

def dispatch_stats_event(stats: StatsData) -> None: ...
def dispatch_game_start_event(data: GameStartData) -> None: ...
def dispatch_round_end_event(data: RoundEndData) -> None: ...
def dispatch_game_end_event(data: GameEndData) -> None: ...
def dispatch_player_death_event(data: DeathData) -> None: ...
def dispatch_team_switch_event(data: TeamSwitchData) -> None: ...

class StatsListener(Thread):
    done: bool
    address: str
    password: str | None
    _in_progress: bool

    def __init__(self) -> None: ...
    def keep_receiving(self) -> None: ...
    def run(self) -> None: ...
    def stop(self) -> None: ...
